# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor.wrapt import ObjectProxy

from contrast.agent import scope
from contrast.agent.assess.policy import propagation_policy
from contrast.agent.assess.policy.source_policy import apply_stream_source
from contrast.utils.decorators import fail_quietly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class BaseStreamProxy(ObjectProxy):
    """
    Base class for implementing stream proxies (marking streams as sources).

    Methods are represented as attributes on the proxy itself which enables us to
    "overwrite" methods even on objects that can't be monkey-patched.

    Proxying a stream (as opposed to instrumenting it directly) ensures that we'll
    be able to modify all relevant attributes. We don't necessarily know the type of
    the stream object we're instrumenting, and sometimes its attributes are
    read-only.

    We also can't use a simple class here, because some frameworks (WebOb) set
    additional attributes on wsgi.input, such as `seek`. We want to maintain as much
    of the original object's functionality as possible, so the best way is with an
    object proxy.

    See https://wrapt.readthedocs.io/en/latest/wrappers.html#proxy-object-attributes
    for details about object property attributes.

    TODO: PYT-917 In the future we may want to instrument write and writelines. This
        would require us to treat the stream more like a propagator than a source
        because trusted data could be written to it.

    TODO: PYT-917 We may eventually need to consider the less common __iter__ method,
        which is currently not instrumented. Werkzeug (at least) does not use __iter__
        when parsing wsgi.input; we may have to investigate WebOb's parser as well. It's
        also possible that the underlying implementation of __iter__ calls an already
        instrumented method for BytesIO, StringIO, etc.

    We might eventually be able to align this proxy hierarchy more closely with
    https://docs.python.org/3/library/io.html#class-hierarchy.
    Note that file objects in python 2 do not line up with this inheritance pattern:
    https://docs.python.org/2/library/stdtypes.html#bltin-file-objects.
    """

    CS__METHOD_NAMES = ["read", "readline", "readlines"]

    read = None
    readline = None
    readlines = None

    cs__tracked = False
    cs__source = False
    cs__properties = None
    cs__source_event = None
    cs__source_type = None
    cs__source_tags = None

    def __init__(self, wrapped):
        super().__init__(wrapped)
        for method_name in self.CS__METHOD_NAMES:
            setattr(self, method_name, getattr(wrapped, method_name, None))


class BaseRawIOProxy(BaseStreamProxy):
    """
    Like BaseStremProxy, but designed for streams that inherit from RawIOBase.

    TODO: PYT-917 We may eventually want to instrument `readinto(b)`
    """

    CS__METHOD_NAMES = BaseStreamProxy.CS__METHOD_NAMES + ["readall"]
    readall = None


class ContrastWsgiStreamProxy(BaseStreamProxy):
    """
    A concrete proxy that marks a PEP-333 compliant wsgi.input stream as a source.

    See https://www.python.org/dev/peps/pep-0333/#input-and-error-streams for the
    specification of this stream.
    """

    def __init__(self, wrapped):
        super().__init__(wrapped)
        instrument_source_stream(self)


class ContrastFileProxy(BaseStreamProxy):
    """
    A concrete proxy that marks a file object (py27 exclusive) as a source.

    This class is identical to ContrastWsgiStreamProxy; however, this is because
    the py27 file object class has a nearly identical spec to the wsgi.input stream.
    """

    def __init__(self, wrapped):
        super().__init__(wrapped)
        instrument_source_stream(self)


class ContrastBufferedReaderProxy(BaseStreamProxy):
    """
    A concrete proxy for BufferedReader (and BufferedRandom) streams.
    This has the same API as WSGI streams and PY2 file objects, but for clarity and
    consistency we've made a unique wrapper.
    """

    def __init__(self, wrapped):
        super().__init__(wrapped)
        instrument_source_stream(self)


class ContrastRawIOProxy(BaseRawIOProxy):
    """
    A concrete proxy that marks a RawIOBase stream as a source.
    """

    def __init__(self, wrapped):
        super().__init__(wrapped)
        instrument_source_stream(self)


def stream_patch_body(method_name, orig_func, self, *args, **kwargs):
    # Since we need to make reference to the input multiple times, convert the
    # first argument to a list and use that instead. This prevents any iterators
    # from being exhausted before we can make use of them in propagation.
    if method_name == "writelines" and len(args):
        args_list = [list(args[0])] + list(args[1:])
    else:
        args_list = args

    result = orig_func(*args_list, **kwargs)

    if scope.in_scope():
        return result

    try:
        with scope.propagation_scope():
            propagation_policy.propagate_stream(
                method_name,
                result,
                self,
                result,
                args_list,
                kwargs,
            )
    except Exception as ex:
        logger.debug("failed to propagate %s", method_name, exc_info=ex)
    finally:
        return result


@fail_quietly("Error instrumenting source stream")
def instrument_source_stream(stream):
    """
    Mark the provided stream as a source, and instrument methods that produce tracked
    content. This must be a stream that we have proxied. i.e., it has the
    `CS__METHOD_NAMES` attribute.

    @param stream: stream object proxied by BaseStreamProxy (must have the
        CS__METHOD_NAMES attribute)
    """
    stream.cs__source = True
    for method_name in stream.CS__METHOD_NAMES:
        _instrument_source_stream_method(stream, method_name)


def _instrument_source_stream_method(stream, method_name):
    """
    Instrument a method on a stream object.

    Only this particular stream instance is instrumented, because the patch is not
    applied at the class level.

    @param stream: file-like object conforming to to PEP 333's wsgi.input specification
    @param method_name: string, the name of the method to instrument
    """
    orig_method = getattr(stream, method_name, None)
    if orig_method is None:
        logger.debug(
            "WARNING: source stream missing expected attribute: %s", method_name
        )
        return
    new_method = _build_patch(stream, orig_method)
    setattr(stream, method_name, new_method)


def _build_patch(stream, orig_method):
    """
    Create a patch for the provided stream method. This is an instance method, so it is
    bound to a specific stream instance.

    Note that we cannot pass orig_method.__self__ to apply_stream_source. This is
    because if we're instrumenting a proxied stream, orig_method.__self__ points to the
    wrapped instance, not the wrapper. Only the wrapper has the cs__* attributes that we
    need for source application.

    @param orig_method: original instance method object
    @return method object that is an instrumented version of the original
    """

    def patched_method(*args, **kwargs):
        result = orig_method(*args, **kwargs)
        if not scope.in_scope():
            apply_stream_source(
                orig_method.__name__, result, stream, result, args, kwargs
            )
        return result

    patched_method.__name__ = orig_method.__name__
    return patched_method



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

from contrast.agent import scope
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import registry
from contrast.agent.policy.applicator import register_import_hooks
from contrast.agent.policy.rewriter import apply_rewrite_policy
from contrast.agent.settings import Settings
from contrast.assess_extensions import cs_str
from contrast.patches import (
    cs_str as cs_str_patches,
    register_assess_patches,
    register_common_patches,
    register_library_patches,
)
from contrast.utils.namespace import Namespace
from contrast.utils.patch_utils import repatch_imported_modules
from contrast_rewriter import register as register_rewriter


HAS_FUNCHOOK = cs_str.has_funchook()
ALWAYS_PURE_PYTHON_METHODS = ["format", "__repr__", "__getitem__"]


class module(Namespace):
    hook = None
    assess_patches_enabled: bool = False


hook_module_map = {
    str: cs_str.create_unicode_hook_module() if HAS_FUNCHOOK else None,
    bytes: cs_str.create_bytes_hook_module() if HAS_FUNCHOOK else None,
    bytearray: cs_str.create_bytearray_hook_module() if HAS_FUNCHOOK else None,
}


def apply_extension_hook(funchook, strtype, hook_module, method_name):
    """
    Apply extension hook for a particular string method

    Returns False if the hook failed to apply and True otherwise
    """
    hook_name = f"apply_{method_name}_hook"
    hook_method = getattr(hook_module, hook_name)

    try:
        hook_method(funchook)
    except RuntimeError as e:
        logger.debug(
            "Failed to apply C extension hook for %s.%s. Will retry in pure python. %s",
            strtype.__name__,
            method_name,
            e,
        )
        return False

    return True


def enable_method_hooks(funchook, hook_module_map, use_extension_hooks=True):
    """
    Enable string propagation hooks for individual string type methods

    This method uses policy to determine all of the string methods that need to be
    patched. For each string type, it first attempts to apply the patch using C
    extension hooks (if not explicitly disallowed by `use_extension_hooks`). If that
    fails for any reason, it will then fall back to applying a pure Python patch.
    """
    use_extension_hooks = use_extension_hooks and HAS_FUNCHOOK

    for strtype, hook_module in hook_module_map.items():
        for node in registry.get_string_method_nodes():
            method_name = node.method_name

            if method_name.lower() in ["cast", "concat"]:
                # these are applied directly in the C extension only
                continue

            real_method_name = (
                "format_map" if method_name == "formatmap" else method_name
            )

            if not hasattr(strtype, real_method_name):
                continue

            if method_name not in ALWAYS_PURE_PYTHON_METHODS and use_extension_hooks:
                if apply_extension_hook(funchook, strtype, hook_module, method_name):
                    logger.debug(
                        "Applied C extension hook for %s.%s",
                        strtype.__name__,
                        method_name,
                    )
                    continue

            cs_str_patches.patch_strtype_method(strtype, real_method_name)
            logger.debug(
                "Applied pure Python patch for %s.%s", strtype.__name__, method_name
            )


@scope.with_contrast_scope
def enable_assess_patches(use_extension_hooks=True):
    """
    Enables extension hooks and other string patches.

    Has no effect if these patches are already enabled.
    """
    if module.assess_patches_enabled:
        return

    logger.debug("has funchook: %s", HAS_FUNCHOOK)

    # NOTE: This function *must* be called before the extension is initialized
    # string_propagation.build_string_propagator_functions()
    string_propagation.build_string_propagator_functions()

    try:
        module.hook = cs_str.initialize()

        cs_str.enable_required_hooks(module.hook)
    except RuntimeError:
        logger.error(
            "Local python builds on OSX may lead to 'Failed to unprotect memory'"
        )
        logger.error(
            "If this applies to you, try running `contrast-fix-interpreter-permissions`"
        )
        raise

    use_extension_hooks = (
        use_extension_hooks
        and not Settings().config.get("agent.python.assess.use_pure_python_hooks")
        and HAS_FUNCHOOK
    )

    cs_str_patches.enable_str_properties()
    enable_method_hooks(
        module.hook, hook_module_map, use_extension_hooks=use_extension_hooks
    )

    try:
        cs_str.install(module.hook)
    except RuntimeError:
        logger.error(
            "Local python builds on OSX may lead to 'Failed to unprotect memory'"
        )
        logger.error(
            "If this applies to you, try running `contrast-fix-interpreter-permissions`"
        )
        raise

    module.assess_patches_enabled = True


def disable_assess_patches():
    """
    Disables extension hooks and other string patches.

    Has no effect if these patches are not already enabled.
    """
    if not module.assess_patches_enabled:
        return

    cs_str.disable(module.hook)

    # Disable any string patches that were applied through Python
    cs_str_patches.unpatch_strtype_methods()

    module.assess_patches_enabled = False


def _enable_protect_patches():
    register_common_patches()

    logger.debug("adding protect policy")
    register_import_hooks(protect_mode=True)

    # This has no effect if the patches are not enabled
    disable_assess_patches()


def _enable_assess_patches(settings: Settings):
    enable_assess_patches()

    # Policy-based rewrites need to be applied prior to any policy patches.
    # Policy patches can be layered on top of rewritten functions. So that
    # means we need to make sure that the "original" function called by the
    # policy patch is the *rewritten* one.
    # Pathlib rewrites must not be applied here. They are only stable when applied by
    # the runner. The exact reason for this is unknown, but it's related to repatching.
    # If this causes noticeable issues, use the runner. This will eventually be
    # mandatory anyway, and policy-based rewrites will be removed from here entirely.
    apply_rewrite_policy(rewrite_pathlib=False)

    logger.debug("enabled assess string patches")
    register_common_patches()
    register_assess_patches()

    logger.debug("adding assess policy")
    register_import_hooks()

    # This is included as a fallback so that we continue to support the case
    # where the runner is not used. If the rewriter has already been enabled by
    # the runner, this has no effect (other than to log a message).
    # Eventually once the runner is fully supported we might feel comfortable
    # removing this code.
    if settings.is_rewriter_enabled:
        register_rewriter(override_config=True)


def _enable_patches():
    settings = Settings()

    if settings.is_analyze_libs_enabled():
        register_library_patches()

    if settings.is_protect_enabled():
        _enable_protect_patches()

    if settings.is_assess_enabled():
        _enable_assess_patches(settings)

    logger.debug("revisiting imported modules to apply patches")
    repatch_imported_modules()


def enable_patches():
    """
    Enable all patches for agent based on current settings
    """
    # Being in scope here prevents us from inadvertently propagating while we are
    # applying patches and navigating policy. This has a fairly significant performance
    # impact for assess initialization, and also promotes correctness/safety.
    with scope.contrast_scope():
        _enable_patches()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast_rewriter import rewrite_for_pytest
from contrast.agent.patch_controller import enable_assess_patches
from contrast.agent.policy.rewriter import apply_rewrite_policy
from contrast.configuration import AgentConfig
from contrast.patches import register_middleware_patches
from contrast_vendor import structlog as logging

from contrast.scripts.runner import USING_RUNNER_ENV_VAR

logger = logging.getLogger("contrast")


def is_runner_in_use() -> bool:
    return os.environ.get(USING_RUNNER_ENV_VAR, "false") == "true"


def start_runner():
    if rewrite_for_pytest():
        if bool(os.environ.get("__POLICY_REWRITES_WHILE_TESTING", False)):
            apply_rewrite_policy(override_config=True)
        return

    config = AgentConfig()

    # Policy-based rewrites need to be applied prior to any policy patches.
    # Policy patches can be layered on top of rewritten functions. So that
    # means we need to make sure that the "original" function called by the
    # policy patch is the *rewritten* one.
    if config.should_apply_policy_rewrites:
        apply_rewrite_policy()

    if config.enable_automatic_middleware:
        register_middleware_patches()

    if config.assess_enabled:
        # NOTE: policy is currently loaded/registered upon import
        from contrast import policy  # pylint: disable=unused-import

        enable_assess_patches()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
This module contains workarounds for the fact that some builtin methods appear
to be unpatchable using funchook for one reason or another.
"""
import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager, registry
from contrast.agent.assess.policy.propagation_policy import (
    PROPAGATOR_ACTIONS,
    track_copy_without_new_event,
)
from contrast.agent.assess.policy.propagators import FormatPropagator, JoinPropagator
from contrast.agent.assess.policy.preshift import Preshift
from contrast.utils.patch_utils import wrap_and_watermark
from contrast.assess_extensions import cs_str, smart_setattr


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_bytearray_join_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def bytearray_join_patch(wrapped, instance, args, kwargs):
        # Since we need to make reference to the input multiple times, convert the
        # first argument to a list and use that instead. This prevents any iterators
        # from being exhausted before we can make use of them in propagation.
        # For bytearray.join, args == (list_or_iterator_of_things_to_join,...)
        # Note that this is different from the C hooks for other join methods. In
        # those cases, the PyObject *args argument corresponds to just the list or
        # iterator itself, in contrast to a tuple that contains that list or
        # iterator. (Got that straight?)
        if cs_str.in_contrast_or_propagation_scope():
            return wrapped(*args, **kwargs)

        args_list = [list(args[0])] + list(args[1:])
        result = wrapped(*args_list, **kwargs)

        scope.enter_contrast_scope()

        try:
            preshift = Preshift(instance, args_list, kwargs)
            propagator = JoinPropagator(node, preshift, result)
            if propagator.needs_propagation:
                propagator.track_and_propagate(result)
        except Exception as ex:
            logger.debug("failed to propagate bytearray.join", exc_info=ex)

        scope.exit_contrast_scope()

        return result

    return wrap_and_watermark(orig_method, bytearray_join_patch)


def build_strtype_join_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def strtype_join_patch(wrapped, instance, args, kwargs):
        if cs_str.in_contrast_or_propagation_scope():
            return wrapped(*args, **kwargs)

        args_list = list(args[0])

        with scope.propagation_scope():
            result = wrapped(args_list, **kwargs)

        scope.enter_contrast_scope()

        try:
            preshift = Preshift(instance, (args_list,), kwargs)
            propagator = JoinPropagator(node, preshift, result)
            if propagator.needs_propagation:
                propagator.track_and_propagate(result)
        except Exception as ex:
            logger.debug("failed to propagate join", exc_info=ex)

        scope.exit_contrast_scope()

        return result

    return wrap_and_watermark(orig_method, strtype_join_patch)


def build_str_format_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def str_format_patch(wrapped, instance, args, kwargs):
        """
        Propagation hook for str.format

        This hook is a special case because we need to enable some propagation to occur
        while we evaluate whether to propagate this particular event. With the current
        general hook infrastructure, this is not possible, so we need to account for it
        here. Eventually it may be possible to fit this back into the more general
        infrastructure if we overhaul the way that scope works.
        """
        result = wrapped(*args, **kwargs)

        if cs_str.in_contrast_or_propagation_scope():
            return result

        try:
            with scope.contrast_scope():
                preshift = Preshift(instance, args, kwargs)
                propagator = FormatPropagator(node, preshift, result)

            # This evaluation must not occur in scope. This is what enables us
            # to perform any conversions from object to __str__ or __repr__,
            # while allowing propagation to occur through those methods if
            # necessary.
            if propagator.needs_propagation:
                with scope.contrast_scope():
                    propagator.track_and_propagate(result)
        except Exception as ex:
            with scope.propagation_scope():
                logger.debug("failed to propagate str.format", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, str_format_patch)


def build_str_formatmap_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]

    def str_formatmap_patch(wrapped, instance, args, kwargs):
        """
        Propagation hook for str.format_map

        This hook is a special case because we need to enable some propagation to occur
        while we evaluate whether to propagate this particular event. With the current
        general hook infrastructure, this is not possible, so we need to account for it
        here. Eventually it may be possible to fit this back into the more general
        infrastructure if we overhaul the way that scope works.
        """
        result = wrapped(*args, **kwargs)

        if cs_str.in_contrast_or_propagation_scope() or not args:
            return result

        try:
            with scope.contrast_scope():
                preshift = Preshift(instance, (), args[0])
                propagator = FormatPropagator(node, preshift, result)

            # This evaluation must not occur in scope. This is what enables us
            # to perform any conversions from object to __str__ or __repr__,
            # while allowing propagation to occur through those methods if
            # necessary.
            if propagator.needs_propagation:
                with scope.contrast_scope():
                    propagator.track_and_propagate(result)
        except Exception as ex:
            with scope.propagation_scope():
                logger.debug("failed to propagate str.format", exc_info=ex)

        return result

    return wrap_and_watermark(orig_method, str_formatmap_patch)


def build_generic_strtype_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]
    propagator_class = PROPAGATOR_ACTIONS.get(node.action)

    def str_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        # This special case does not apply to bytearrays
        if result is instance:
            return result

        if cs_str.in_contrast_or_propagation_scope():
            return result

        scope.enter_contrast_scope()

        try:
            preshift = Preshift(instance, args, kwargs)
            propagator = propagator_class(node, preshift, result)

            if propagator.needs_propagation:
                propagator.track_and_propagate(result)
        except Exception as ex:
            name = orig_method.__class__.__name__
            logger.debug(
                "failed to propagate %s.%s", name, orig_method.__name__, exc_info=ex
            )

        scope.exit_contrast_scope()

        return result

    return wrap_and_watermark(orig_method, str_patch)


def build_track_without_new_event_patch(orig_method, patch_policy):
    node = patch_policy.propagator_nodes[0]
    propagator_class = PROPAGATOR_ACTIONS.get(node.action)

    def str_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        if cs_str.in_contrast_or_propagation_scope():
            return result

        # This special case applies to all bytearray methods and all .translate methods
        if result == instance:
            track_copy_without_new_event(result, instance)
            return result

        scope.enter_contrast_scope()

        try:
            preshift = Preshift(instance, args, kwargs)
            propagator = propagator_class(node, preshift, result)

            if propagator.needs_propagation:
                propagator.track_and_propagate(result)
        except Exception as ex:
            name = orig_method.__class__.__name__
            logger.debug(
                "failed to propagate %s.%s", name, orig_method.__name__, exc_info=ex
            )

        scope.exit_contrast_scope()

        return result

    return wrap_and_watermark(orig_method, str_patch)


def build_and_apply_patch(owner, method_name, patch_builder):
    orig_method = getattr(owner, method_name)

    policy_method_name = (
        "formatmap" if orig_method.__name__ == "format_map" else orig_method.__name__
    )

    patch_policy = registry.get_policy_by_name("BUILTIN.str." + policy_method_name)
    patch = patch_builder(orig_method, patch_policy)

    patch_manager.patch(owner, method_name, patch)


def property_getter(self):
    return contrast.STRING_TRACKER.get(self, None)


def property_setter(self, value):
    contrast.STRING_TRACKER.update_properties(self, value)


def enable_str_properties():
    strprop = property(fget=property_getter, fset=property_setter)

    smart_setattr(str, "cs__properties", strprop)
    smart_setattr(bytes, "cs__properties", strprop)
    smart_setattr(bytearray, "cs__properties", strprop)


def patch_strtype_method(strtype, method_name):
    if method_name == "join":
        builder = (
            build_bytearray_join_patch
            if strtype is bytearray
            else build_strtype_join_patch
        )
    elif method_name == "format":
        builder = build_str_format_patch
    elif method_name == "format_map":
        builder = build_str_formatmap_patch
    elif method_name == "translate" and strtype is str:
        builder = build_track_without_new_event_patch
    else:
        builder = (
            build_track_without_new_event_patch
            if strtype is bytearray
            else build_generic_strtype_patch
        )

    build_and_apply_patch(strtype, method_name, patch_builder=builder)


def unpatch_strtype_methods():
    """
    Replace all patched strtype methods with the original implementation
    """
    patch_manager.reverse_patches_by_owner(str)
    patch_manager.reverse_patches_by_owner(bytes)
    patch_manager.reverse_patches_by_owner(bytearray)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Controller for global scope state

Basically we use scoping to prevent us from assessing our own code. Scope
improves performance but it also prevents us from accidentally recursing
inside our analysis code. For example, we don't want to inadvertently cause
string propagation events while we're doing string building for reporting
purposes.
"""
import sys

from contextlib import contextmanager

from contrast.assess_extensions import cs_str


__SCOPE_LEVELS = ("contrast", "propagation", "trigger")


def __build_func(func, level):
    func = getattr(cs_str, func)

    def _func():
        return func(level)

    return _func


def __build_context(enter, exit_):
    def _func():
        enter()
        try:
            yield
        finally:
            exit_()

    return contextmanager(_func)


def __build_decorator(context):
    def decorator(orig_func):
        def wrapper(*args, **kwargs):
            with context():
                return orig_func(*args, **kwargs)

        return wrapper

    return decorator


def __generate_scope_functions(scope_levels):
    """
    Auto-generates scope control API based on scope level names
    """
    mod = sys.modules[__name__]

    for name in scope_levels:
        level = getattr(cs_str, f"{name.upper()}_SCOPE")

        enter = __build_func("enter_scope", level)
        exit_ = __build_func("exit_scope", level)
        in_ = __build_func("in_scope", level)
        context = __build_context(enter, exit_)
        decorator = __build_decorator(context)

        enter.__doc__ = f"Enter {name} scope"
        exit_.__doc__ = f"Exit {name} scope"
        in_.__doc__ = f"Returns True if in {name} scope"
        context.__doc__ = f"Context manager for {name} scope"
        decorator.__doc__ = f"Decorator for {name} scope"

        setattr(mod, f"enter_{name}_scope", enter)
        setattr(mod, f"exit_{name}_scope", exit_)
        setattr(mod, f"in_{name}_scope", in_)
        setattr(mod, f"{name}_scope", context)
        setattr(mod, f"with_{name}_scope", decorator)


__generate_scope_functions(__SCOPE_LEVELS)


enter_contrast_scope.__doc__ = """
    Enter contrast scope

    Contrast scope is global. It should prevent us from taking *any*
    further analysis action, whether it be propagation or evaluating
    triggers.
"""
enter_propagation_scope.__doc__ = """
    Enter propagation scope
     
    While in propagation scope, prevent any further propagation actions.
    Basically this means that no string propagation should occur while in
    propagation scope.
"""
enter_trigger_scope.__doc__ = """
    Enter trigger scope

    While in trigger scope, prevent analysis inside of any other trigger
    methods that get called.
"""


def in_scope():
    """Indicates we are in either contrast scope or propagation scope"""
    return cs_str.in_contrast_or_propagation_scope()


@contextmanager
def pop_contrast_scope():
    """
    Context manager that pops contrast scope level and restores it when it exits

    Scope is implemented as a stack. If the thread is in contrast scope at the time
    this is called, the scope level will be reduced by one for the lifetime of the
    context manager. If the prior scope level was 1, this has the effect of temporarily
    disabling contrast scope. The original scope level will be restored when the
    context manager exits. If the thread is **not** already in contrast scope when this
    is called, it has no effect.
    """
    in_scope = in_contrast_scope()
    # This has no effect if we're not already in scope
    exit_contrast_scope()
    try:
        yield
    finally:
        # For safety, only restore scope if we were in it to begin with
        if in_scope:
            enter_contrast_scope()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import copy_from


class KeepPropagator(BasePropagator):
    def _propagate(self):
        self.first_source = self.sources[0]
        copy_from(self.target, self.first_source, 0, self.node.untags)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Helpers that make development with *structlog* more pleasant.

See also the narrative documentation in `development`.
"""

from __future__ import annotations

import sys
import warnings

from io import StringIO
from typing import Any, Iterable, TextIO, Type, Union

from ._frames import _format_exception
from .processors import _figure_out_exc_info
from .typing import EventDict, ExceptionRenderer, ExcInfo, WrappedLogger


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from contrast_vendor.typing_extensions import Protocol


try:
    import colorama
except ImportError:
    colorama = None

try:
    import better_exceptions
except ImportError:
    better_exceptions = None

try:
    import rich

    from rich.console import Console
    from rich.traceback import Traceback
except ImportError:
    rich = None  # type: ignore[assignment]


__all__ = [
    "ConsoleRenderer",
    "plain_traceback",
    "rich_traceback",
    "better_traceback",
]

_IS_WINDOWS = sys.platform == "win32"

_MISSING = "{who} requires the {package} package installed.  "
_EVENT_WIDTH = 30  # pad the event name to so many characters


def _pad(s: str, length: int) -> str:
    """
    Pads *s* to length *length*.
    """
    missing = length - len(s)

    return s + " " * (missing if missing > 0 else 0)


if colorama is not None:
    RESET_ALL = colorama.Style.RESET_ALL
    BRIGHT = colorama.Style.BRIGHT
    DIM = colorama.Style.DIM
    RED = colorama.Fore.RED
    BLUE = colorama.Fore.BLUE
    CYAN = colorama.Fore.CYAN
    MAGENTA = colorama.Fore.MAGENTA
    YELLOW = colorama.Fore.YELLOW
    GREEN = colorama.Fore.GREEN
    RED_BACK = colorama.Back.RED
else:
    # These are the same values as the Colorama color codes. Redefining them
    # here allows users to specify that they want color without having to
    # install Colorama, which is only supposed to be necessary in Windows.
    RESET_ALL = "\033[0m"
    BRIGHT = "\033[1m"
    DIM = "\033[2m"
    RED = "\033[31m"
    BLUE = "\033[34m"
    CYAN = "\033[36m"
    MAGENTA = "\033[35m"
    YELLOW = "\033[33m"
    GREEN = "\033[32m"
    RED_BACK = "\033[41m"


if _IS_WINDOWS:  # pragma: no cover
    # On Windows, use colors by default only if Colorama is installed.
    _has_colors = colorama is not None
else:
    # On other OSes, use colors by default.
    _has_colors = True

# Prevent breakage of packages that used the old name of the variable.
_use_colors = _has_colors


class _Styles(Protocol):
    reset: str
    bright: str
    level_critical: str
    level_exception: str
    level_error: str
    level_warn: str
    level_info: str
    level_debug: str
    level_notset: str

    timestamp: str
    logger_name: str
    kv_key: str
    kv_value: str


Styles = Union[_Styles, Type[_Styles]]


class _ColorfulStyles:
    reset = RESET_ALL
    bright = BRIGHT

    level_critical = RED
    level_exception = RED
    level_error = RED
    level_warn = YELLOW
    level_info = GREEN
    level_debug = GREEN
    level_notset = RED_BACK

    timestamp = DIM
    logger_name = BLUE
    kv_key = CYAN
    kv_value = MAGENTA


class _PlainStyles:
    reset = ""
    bright = ""

    level_critical = ""
    level_exception = ""
    level_error = ""
    level_warn = ""
    level_info = ""
    level_debug = ""
    level_notset = ""

    timestamp = ""
    logger_name = ""
    kv_key = ""
    kv_value = ""


def plain_traceback(sio: TextIO, exc_info: ExcInfo) -> None:
    """
    "Pretty"-print *exc_info* to *sio* using our own plain formatter.

    To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

    Used by default if neither *Rich* not *better-exceptions* are present.

    .. versionadded:: 21.2
    """
    sio.write("\n" + _format_exception(exc_info))


def rich_traceback(sio: TextIO, exc_info: ExcInfo) -> None:
    """
    Pretty-print *exc_info* to *sio* using the *Rich* package.

    To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

    Used by default if *Rich* is installed.

    .. versionadded:: 21.2
    """
    sio.write("\n")
    Console(file=sio, color_system="truecolor").print(
        Traceback.from_exception(*exc_info, show_locals=True)
    )


def better_traceback(sio: TextIO, exc_info: ExcInfo) -> None:
    """
    Pretty-print *exc_info* to *sio* using the *better-exceptions* package.

    To be passed into `ConsoleRenderer`'s ``exception_formatter`` argument.

    Used by default if *better-exceptions* is installed and *Rich* is absent.

    .. versionadded:: 21.2
    """
    sio.write("\n" + "".join(better_exceptions.format_exception(*exc_info)))


if rich is not None:
    default_exception_formatter = rich_traceback
elif better_exceptions is not None:
    default_exception_formatter = better_traceback
else:
    default_exception_formatter = plain_traceback


class ConsoleRenderer:
    """
    Render ``event_dict`` nicely aligned, possibly in colors, and ordered.

    If ``event_dict`` contains a true-ish ``exc_info`` key, it will be
    rendered *after* the log line. If Rich_ or better-exceptions_ are present,
    in colors and with extra context.

    :param pad_event: Pad the event to this many characters.
    :param colors: Use colors for a nicer output. `True` by default. On
        Windows only if Colorama_ is installed.
    :param force_colors: Force colors even for non-tty destinations.
        Use this option if your logs are stored in a file that is meant
        to be streamed to the console. Only meaningful on Windows.
    :param repr_native_str: When `True`, `repr` is also applied
        to native strings (i.e. unicode on Python 3 and bytes on Python 2).
        Setting this to `False` is useful if you want to have human-readable
        non-ASCII output on Python 2.  The ``event`` key is *never*
        `repr` -ed.
    :param level_styles: When present, use these styles for colors. This
        must be a dict from level names (strings) to Colorama styles. The
        default can be obtained by calling
        `ConsoleRenderer.get_default_level_styles`
    :param exception_formatter: A callable to render ``exc_infos``. If rich_
        or better-exceptions_ are installed, they are used for pretty-printing
        by default (rich_ taking precedence). You can also manually set it to
        `plain_traceback`, `better_traceback`, `rich_traceback`, or implement
        your own.
    :param sort_keys: Whether to sort keys when formatting. `True` by default.
    :param event_key: The key to look for the main log message. Needed when
        you rename it e.g. using `structlog.processors.EventRenamer`.

    Requires the Colorama_ package if *colors* is `True` **on Windows**.

    .. _Colorama: https://pypi.org/project/colorama/
    .. _better-exceptions: https://pypi.org/project/better-exceptions/
    .. _Rich: https://pypi.org/project/rich/

    .. versionadded:: 16.0
    .. versionadded:: 16.1 *colors*
    .. versionadded:: 17.1 *repr_native_str*
    .. versionadded:: 18.1 *force_colors*
    .. versionadded:: 18.1 *level_styles*
    .. versionchanged:: 19.2
       *Colorama* now initializes lazily to avoid unwanted initializations as
       ``ConsoleRenderer`` is used by default.
    .. versionchanged:: 19.2 Can be pickled now.
    .. versionchanged:: 20.1 *Colorama* does not initialize lazily on Windows
       anymore because it breaks rendering.
    .. versionchanged:: 21.1 It is additionally possible to set the logger name
       using the ``logger_name`` key in the ``event_dict``.
    .. versionadded:: 21.2 *exception_formatter*
    .. versionchanged:: 21.2 `ConsoleRenderer` now handles the ``exc_info``
       event dict key itself. Do **not** use the
       `structlog.processors.format_exc_info` processor together with
       `ConsoleRenderer` anymore! It will keep working, but you can't have
       customize exception formatting and a warning will be raised if you ask
       for it.
    .. versionchanged:: 21.2 The colors keyword now defaults to True on
       non-Windows systems, and either True or False in Windows depending on
       whether *Colorama* is installed.
    .. versionadded:: 21.3.0 *sort_keys*
    .. versionadded:: 22.1 *event_key*
    """

    def __init__(
        self,
        pad_event: int = _EVENT_WIDTH,
        colors: bool = _has_colors,
        force_colors: bool = False,
        repr_native_str: bool = False,
        level_styles: Styles | None = None,
        exception_formatter: ExceptionRenderer = default_exception_formatter,
        sort_keys: bool = True,
        event_key: str = "event",
    ):
        styles: Styles
        if colors:
            if _IS_WINDOWS:  # pragma: no cover
                # On Windows, we can't do colorful output without colorama.
                if colorama is None:
                    classname = self.__class__.__name__
                    raise SystemError(
                        _MISSING.format(
                            who=classname + " with `colors=True`",
                            package="colorama",
                        )
                    )
                # Colorama must be init'd on Windows, but must NOT be
                # init'd on other OSes, because it can break colors.
                if force_colors:
                    colorama.deinit()
                    colorama.init(strip=False)
                else:
                    colorama.init()

            styles = _ColorfulStyles
        else:
            styles = _PlainStyles

        self._styles = styles
        self._pad_event = pad_event

        if level_styles is None:
            self._level_to_color = self.get_default_level_styles(colors)
        else:
            self._level_to_color = level_styles

        for key in self._level_to_color:
            self._level_to_color[key] += styles.bright
        self._longest_level = len(
            max(self._level_to_color.keys(), key=lambda e: len(e))
        )

        self._repr_native_str = repr_native_str
        self._exception_formatter = exception_formatter
        self._sort_keys = sort_keys
        self._event_key = event_key

    def _repr(self, val: Any) -> str:
        """
        Determine representation of *val* depending on its type &
        self._repr_native_str.
        """
        if self._repr_native_str is True:
            return repr(val)

        if isinstance(val, str):
            return val

        return repr(val)

    def __call__(
        self, logger: WrappedLogger, name: str, event_dict: EventDict
    ) -> str:
        sio = StringIO()

        ts = event_dict.pop("timestamp", None)
        if ts is not None:
            sio.write(
                # can be a number if timestamp is UNIXy
                self._styles.timestamp
                + str(ts)
                + self._styles.reset
                + " "
            )
        level = event_dict.pop("level", None)
        if level is not None:
            sio.write(
                "["
                + self._level_to_color.get(level, "")
                + _pad(level, self._longest_level)
                + self._styles.reset
                + "] "
            )

        # force event to str for compatibility with standard library
        event = event_dict.pop(self._event_key, None)
        if not isinstance(event, str):
            event = str(event)

        if event_dict:
            event = _pad(event, self._pad_event) + self._styles.reset + " "
        else:
            event += self._styles.reset
        sio.write(self._styles.bright + event)

        logger_name = event_dict.pop("logger", None)
        if logger_name is None:
            logger_name = event_dict.pop("logger_name", None)

        if logger_name is not None:
            sio.write(
                "["
                + self._styles.logger_name
                + self._styles.bright
                + logger_name
                + self._styles.reset
                + "] "
            )

        stack = event_dict.pop("stack", None)
        exc = event_dict.pop("exception", None)
        exc_info = event_dict.pop("exc_info", None)

        event_dict_keys: Iterable[str] = event_dict.keys()
        if self._sort_keys:
            event_dict_keys = sorted(event_dict_keys)

        sio.write(
            " ".join(
                self._styles.kv_key
                + key
                + self._styles.reset
                + "="
                + self._styles.kv_value
                + self._repr(event_dict[key])
                + self._styles.reset
                for key in event_dict_keys
            )
        )

        if stack is not None:
            sio.write("\n" + stack)
            if exc_info or exc is not None:
                sio.write("\n\n" + "=" * 79 + "\n")

        if exc_info:
            exc_info = _figure_out_exc_info(exc_info)

            self._exception_formatter(sio, exc_info)
        elif exc is not None:
            if self._exception_formatter is not plain_traceback:
                warnings.warn(
                    "Remove `format_exc_info` from your processor chain "
                    "if you want pretty exceptions.",
                    stacklevel=2,
                )
            sio.write("\n" + exc)

        return sio.getvalue()

    @staticmethod
    def get_default_level_styles(colors: bool = True) -> Any:
        """
        Get the default styles for log levels

        This is intended to be used with `ConsoleRenderer`'s ``level_styles``
        parameter.  For example, if you are adding custom levels in your
        home-grown :func:`~structlog.stdlib.add_log_level` you could do::

            my_styles = ConsoleRenderer.get_default_level_styles()
            my_styles["EVERYTHING_IS_ON_FIRE"] = my_styles["critical"]
            renderer = ConsoleRenderer(level_styles=my_styles)

        :param colors: Whether to use colorful styles. This must match the
            *colors* parameter to `ConsoleRenderer`. Default: `True`.
        """
        styles: Styles
        styles = _ColorfulStyles if colors else _PlainStyles
        return {
            "critical": styles.level_critical,
            "exception": styles.level_exception,
            "error": styles.level_error,
            "warn": styles.level_warn,
            "warning": styles.level_warn,
            "info": styles.level_info,
            "debug": styles.level_debug,
            "notset": styles.level_notset,
        }


_SENTINEL = object()


def set_exc_info(
    logger: WrappedLogger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Set ``event_dict["exc_info"] = True`` if *method_name* is ``"exception"``.

    Do nothing if the name is different or ``exc_info`` is already set.
    """
    if (
        method_name != "exception"
        or event_dict.get("exc_info", _SENTINEL) is not _SENTINEL
    ):
        return event_dict

    event_dict["exc_info"] = True

    return event_dict



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .finding import Finding, send_finding
from .library import Library
from .route_coverage import Route
from .type_checked_property import TypeCheckedProperty



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import gc
import inspect

from contrast.assess_extensions.cs_str import has_funchook
from contrast.agent import scope
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import patch_manager
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.utils.namespace import Namespace


class module(Namespace):
    patched_classes = []


def find_str_subclasses():
    """
    Digs through the garbage (collector) to find any subclasses of str
    """
    for obj in gc.get_objects():
        try:
            if inspect.isclass(obj) and issubclass(obj, (str, bytes, bytearray)):
                yield obj
        except Exception:
            continue


@fail_quietly("Failed to propagate string subclass __new__")
@scope.with_contrast_scope
def propagate_str_cast(result, args, kwargs):
    string_propagation.propagate_unicode_cast(result, args[0], result, args[1:], kwargs)


@fail_quietly("Failed to propagate bytes subclass __new__")
@scope.with_contrast_scope
def propagate_bytes_cast(result, args, kwargs):
    string_propagation.propagate_bytes_cast(result, args[0], result, args[1:], kwargs)


@fail_quietly("Failed to propagate bytearray subclass __init__")
@scope.with_contrast_scope
def propagate_bytearray_init(instance, args, kwargs):
    """
    To make use of our generalized bytearray cast propagation machinery, we need to call
    `propagate_bytearray_cast` as though this were a __new__ patch. The calling
    semantics of __new__ differ from __init__, so we shuffle some things around
    accordingly.
    """
    target = instance
    result = instance
    self_obj = None
    string_propagation.propagate_bytearray_cast(target, result, self_obj, args, kwargs)


def build_cast_patch(orig_func, patch_policy, propagation_func):
    del patch_policy

    def __new__(*args, **kwargs):
        result = orig_func(*args, **kwargs)
        propagation_func(result, args, kwargs)
        return result

    # NOTE: builtin methods can't be wrapped
    return __new__


def build_bytearray_init_patch(orig_func, patch_policy):
    del patch_policy

    def __init__(self, *args, **kwargs) -> None:
        result = orig_func(self, *args, **kwargs)
        propagate_bytearray_init(self, args, kwargs)
        return result

    return __init__


@fail_quietly("Failed to apply patches for string subclasses")
def register_patches():
    """
    Applies patches to known subclasses of str

    Problem statement: the cast/new hook for subclasses of str does not work
    without funchook. This seems to be an issue with timing: subclasses that
    are defined *after* our (non-funchook) extension hooks are applied do not
    appear to have a problem.

    However, even with the runner, our instrumentation is not necessarily going
    to be early enough to affect *all* subclasses that are defined.

    The solution is to use the garbage collector to find any subclasses of str
    that may already be defined at this point in time. We then explicitly apply
    propagation patches to the __new__ methods of these classes.
    """
    if has_funchook():
        return

    for cls in find_str_subclasses():
        if issubclass(cls, bytearray):
            build_and_apply_patch(
                cls,
                "__init__",
                build_bytearray_init_patch,
            )
            module.patched_classes.append(cls)
            continue

        propagation_func = None
        if issubclass(cls, str):
            propagation_func = propagate_str_cast
        elif issubclass(cls, bytes):
            propagation_func = propagate_bytes_cast

        if propagation_func is not None:
            build_and_apply_patch(
                cls, "__new__", build_cast_patch, builder_args=(propagation_func,)
            )
            module.patched_classes.append(cls)


def reverse_patches():
    for cls in module.patched_classes:
        patch_manager.reverse_patches_by_owner(cls)

    module.patched_classes = []



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class Singleton(object):
    """
    Base class for children that provide the Singleton pattern

    Singletons are classes that guarantee that only a single instance can ever
    be created. Any time a "new" instance of the class is instantiated, it
    actually refers to the same underlying object. This is useful for providing
    a container for global state.

    This implementation was taken from the Python documentation (admittedly,
    from a very old version):
        https://www.python.org/download/releases/2.2/descrintro/#__new__
    """

    def __new__(cls, *args, **kwds):
        instance = cls.__dict__.get("__instance__")
        if instance is not None:
            return instance

        cls.__instance__ = instance = object.__new__(cls)
        instance.init(*args, **kwds)
        return instance

    def init(self, *args, **kwds):
        """
        Subclasses should override this method for initialization, and not __init__.
        """
        pass

    @classmethod
    def clear_instance(cls):
        """
        Delete the singleton's current instance.

        While singletons are objects intended to have one instance only, there
        may be time when deleting the existing instance is needed.
        """
        instance = cls.__dict__.get("__instance__")
        if instance is not None:
            del cls.__instance__



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import base64

from contrast.utils.string_utils import ensure_string


def base64_encode(to_encode):
    """
    Allows you to encode string to Base64
    """
    if isinstance(to_encode, str):
        input_object = to_encode.encode("utf-8")
    else:
        input_object = to_encode

    output_object = base64.b64encode(input_object)

    return output_object.decode("ascii")


def base64_decode(to_decode):
    """
    Allows you to decode Base64 to str
    """
    output_object = base64.urlsafe_b64decode(ensure_string(to_decode))
    return ensure_string(output_object)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading

import contrast
from contrast.agent.heartbeat_thread import HEARTBEAT_THREAD_NAME
from contrast.agent.server_settings_poll import SERVER_SETTING_THREAD_NAME
from contrast.agent.telemetry import Telemetry, TELEMETRY_THREAD_NAME
from contrast.reporting.reporting_client import (
    ReportingClient,
    REPORTING_CLIENT_THREAD_NAME,
)
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_quietly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

LOG_MSG = "%s thread wasn't running - restarting it now"

# The lock here is to prevent a race condition when restarting background threads.
# We need to ensure that only one thread at once can get the list of running threads and
# act accordingly based on that list.
MODULE_LOCK = threading.Lock()


@fail_quietly("failed to check background threads")
def ensure_running(agent_state, skip_telemetry=False):
    """
    Check that long-running agent background threads are running in the current process.
    Restart any threads that appear to have been killed.

    This occurs most often when a webserver that preloads the application forks its
    master process to spawn workers. In this case, any threads started in the master
    process don't transfer over to workers, so they need to be restarted.

    :param skip_telemetry: If `True`, do not check the telemetry thread
    """
    logger.debug("checking background threads")

    with MODULE_LOCK:
        # PERF: this is a critical section (inside of a lock). Be mindful!
        threads_by_name = {t.name: t for t in threading.enumerate()}
        if not skip_telemetry:
            _check_telemetry(threads_by_name)
        _check_heartbeat(threads_by_name)
        _check_server_settings_poll(threads_by_name)
        _check_reporting_client(threads_by_name, agent_state)


def _check_telemetry(threads_by_name):
    # PERF: this is a critical section (inside of a lock). Be mindful!
    thread = threads_by_name.get(TELEMETRY_THREAD_NAME)
    if thread is not None or contrast.telemetry_disabled():
        return

    logger.debug(LOG_MSG, TELEMETRY_THREAD_NAME)
    contrast.TELEMETRY = Telemetry()
    contrast.TELEMETRY.start()


def _check_heartbeat(threads_by_name):
    # PERF: this is a critical section (inside of a lock). Be mindful!
    thread = threads_by_name.get(HEARTBEAT_THREAD_NAME)
    if thread is not None:
        return

    settings = Settings()
    logger.debug(LOG_MSG, HEARTBEAT_THREAD_NAME)
    settings.heartbeat = None
    settings.establish_heartbeat()


def _check_server_settings_poll(threads_by_name):
    # PERF: this is a critical section (inside of a lock). Be mindful!
    thread = threads_by_name.get(SERVER_SETTING_THREAD_NAME)
    if thread is not None:
        return

    settings = Settings()
    logger.debug(LOG_MSG, SERVER_SETTING_THREAD_NAME)
    settings.server_settings_poll = None
    settings.establish_server_settings_poll()


def _check_reporting_client(threads_by_name, agent_state):
    # PERF: this is a critical section (inside of a lock). Be mindful!

    # something about ReportingClient being a singleton prevents it from being named
    # correctly (sometimes). The attempt to find the thread here is slower but necessary
    # TODO: PYT-1960 Revert this back to lookup by thread name only
    thread = {type(t): t for t in threads_by_name.values()}.get(ReportingClient)
    if thread is not None:
        return

    # the reporting client thread should always be running; there are no conditions here

    logger.debug(LOG_MSG, REPORTING_CLIENT_THREAD_NAME)
    ReportingClient.clear_instance()
    agent_state.reporting_client = ReportingClient()
    agent_state.reporting_client.start()



"""
Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
76 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019 Python Software
77 Foundation; All Rights Reserved
"""


def _calc___package__(globals):
    """Calculate what __package__ should be.

    __package__ is not guaranteed to be defined or could be set to None
    to represent that its proper value is unknown.

    """
    package = globals.get('__package__')
    if package is None:
        package = globals.get('__name__')

        if package is None:
            return package

        if '__path__' not in globals:
            package = package.rpartition('.')[0]
    return package


def _resolve_name(name, package, level):
    """
    Resolve a relative module name to an absolute one.
    """
    bits = package.rsplit('.', level - 1)
    if len(bits) < level:
        return None
    base = bits[0]
    return '{}.{}'.format(base, name) if name else base


def resolve_relative_import_name(name, level, globals):
    """
    Part of this code is ported from importlib._bootstrap._gcd_import in cpython

    The 'globals' argument is used to infer where the import is occurring from
    to handle relative imports.
    """
    abs_name = None

    if level <= 0 or not name or not isinstance(name, str):
        return abs_name

    package = _calc___package__(globals or {})
    if package and isinstance(package, str):
        abs_name = _resolve_name(name, package, level)

    return abs_name




from sys import version_info

from . import hookspy


py_version = version_info[:2]


fastcall_name = "fastcall_func"
fastcall_typedef = (
    "typedef PyObject *(*{})(PyObject *, PyObject *const *, Py_ssize_t)".format(
        fastcall_name
    )
)

fastcall_kwargs_name = "fastcall_kwargs_func"
fastcall_kwargs_typedef = "typedef PyObject *(*{})(PyObject *, PyObject *const *, Py_ssize_t, PyObject *)".format(
    fastcall_kwargs_name
)


unaryfunc_signature = "PyObject *{}(PyObject *self)"
binaryfunc_signature = "PyObject *{}(PyObject *self, PyObject *args)"
ternaryfunc_signature = "PyObject *{}(PyObject *self, PyObject *args, PyObject *kwargs)"
fastcall_signature = (
    "PyObject *{}(PyObject *self, PyObject *const *args, Py_ssize_t nargs)"
)
fastcall_kwargs_signature = "PyObject *{}(PyObject *self, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames)"

newfunc_signature = "PyObject *{}(PyTypeObject *type, PyObject *args, PyObject *kwds)"
initproc_signature = "int {}(PyObject *self, PyObject *args, PyObject *kwds)"

unaryfunc_call = "(self)"
binaryfunc_call = "(self, args)"
ternaryfunc_call = "(self, args, kwargs)"
fastcall_call = "(self, args, nargs)"
fastcall_kwargs_call = "(self, args, nargs, kwargs)"
newfunc_call = "(type, args, kwds)"
initproc_call = "(self, args, kwds)"


signatures_by_name = {
    "unaryfunc": unaryfunc_signature,
    "binaryfunc": binaryfunc_signature,
    "ternaryfunc": ternaryfunc_signature,
    fastcall_name: fastcall_signature,
    fastcall_kwargs_name: fastcall_kwargs_signature,
    "newfunc": newfunc_signature,
    "initproc": initproc_signature,
}


calls_by_name = {
    "unaryfunc": unaryfunc_call,
    "binaryfunc": binaryfunc_call,
    "ternaryfunc": ternaryfunc_call,
    fastcall_name: fastcall_call,
    fastcall_kwargs_name: fastcall_kwargs_call,
    "newfunc": newfunc_call,
    "initproc": initproc_call,
}


propagate_args_by_name = {
    "unaryfunc": "NULL, NULL",
    "binaryfunc": "args, NULL",
    "ternaryfunc": "args, kwargs",
}


method_signatures_map = {
    hookspy.METH_O: "binaryfunc",
    hookspy.METH_NOARGS: "unaryfunc",
    hookspy.METH_VARARGS: "binaryfunc",
    hookspy.METH_KEYWORDS: "ternaryfunc",
    hookspy.METH_FASTCALL: fastcall_name,
    hookspy.METH_VARARGS | hookspy.METH_KEYWORDS: "ternaryfunc",
    hookspy.METH_FASTCALL | hookspy.METH_KEYWORDS: fastcall_kwargs_name,
    # special cases:
    hookspy.METH_NEW: "newfunc",
    hookspy.METH_INIT: "initproc",
}


def is_buggy_method(strtype, name):
    if strtype not in ["bytes", "bytearray"] or py_version != (3, 6):
        return False
    return name in ["decode", "split", "rsplit", "splitlines", "translate"]


def get_method_signature(strtype, name, argtype):
    # There is a bug in 3.6 that requires this special handling logic
    if is_buggy_method(strtype, name):
        return fastcall_kwargs_name
    return method_signatures_map[argtype]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy import propagation_policy
from contrast.agent.assess.policy import source_policy
from contrast.agent.assess.policy import trigger_policy
from contrast.agent.assess.utils import get_self_for_method
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_loudly


def check_or_enter_scope(orig_func):
    """
    Decorator that checks if we're in contrast scope. If so, return immediately, else
    call the original function in contrast scope.

    Sometimes it is essential that the very first action we take when a method is called
    is a scope check. If we hit any instrumented method before this check, we'll be in
    an infinite recursion. This decorator must be the topmost (first) if it's used.
    """

    def wrapper(*args, **kwargs):
        if scope.in_contrast_scope():
            return
        with scope.contrast_scope():
            orig_func(*args, **kwargs)

    return wrapper


@check_or_enter_scope  # NOTE: this decorator must come first!
@fail_loudly("Failed to perform assess analysis.")
def analyze(context, patch_policy, result, args, kwargs):
    if not context or not Settings().is_assess_enabled():
        return

    self_obj = get_self_for_method(patch_policy, args)
    preshift = Preshift(self_obj, args, kwargs)

    _analyze(patch_policy, preshift, self_obj, result, args, kwargs)


def skip_analysis(context):
    """
    Skip analysis if there is no context, scope, or configuration is False
    :param context: RequestContext
    :return:
    """
    if not context:
        return True
    if scope.in_contrast_scope():
        return True
    return not Settings().is_assess_enabled()


def _analyze(patch_policy, preshift, self_obj, ret, orig_args, orig_kwargs=None):
    if not patch_policy:
        return

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    if patch_policy.analysis_func is not None:
        rule = (
            patch_policy.trigger_nodes[0].rule if patch_policy.trigger_nodes else None
        )

        patch_policy.analysis_func(
            rule=rule,
            nodes=patch_policy.all_nodes,
            self_obj=self_obj,
            ret=ret,
            orig_args=orig_args,
            orig_kwargs=orig_kwargs,
            preshift=preshift,
        )
    else:
        _run_unoptimized_analysis(
            patch_policy, preshift, self_obj, ret, orig_args, orig_kwargs
        )


def _run_unoptimized_analysis(patch_policy, preshift, self_obj, ret, args, kwargs):
    if patch_policy.trigger_nodes:
        # Each node may potentially correspond to a different rule
        for node in patch_policy.trigger_nodes:
            trigger_policy.apply(
                rule=node.rule,
                nodes=[node],
                ret=ret,
                orig_args=args,
                orig_kwargs=kwargs,
            )

    if patch_policy.source_nodes:
        source_policy.apply(
            nodes=patch_policy.source_nodes,
            self_obj=self_obj,
            ret=ret,
            orig_args=args,
            orig_kwargs=kwargs,
        )

    if patch_policy.propagator_nodes:
        propagation_policy.apply(
            nodes=patch_policy.propagator_nodes, preshift=preshift, ret=ret
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.settings import Settings

DATABASE_MYSQL = "MySQL"
DATABASE_PG = "PostgreSQL"
DATABASE_SQL_ALCHEMY = "SQLAlchemy"
DATABASE_SQLITE = "SQLite3"

ACTION_EXEC = "exec"
ACTION_EXECUTE = "execute"


def apply_rule(patch_policy, orig_func, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()

    if (
        context is not None
        and Settings().is_protect_enabled()
        and not scope.in_contrast_scope()
    ):
        for node in patch_policy.trigger_nodes:
            rule_name = node.rule.name
            rule = Settings().protect_rules.get(rule_name)
            if not rule or not rule.enabled:
                continue

            for source in node.get_protect_sources(args, kwargs):
                rule.protect(patch_policy, source, args, kwargs)

    return orig_func(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import Enum, auto


class ProtectionRule(Enum):
    OFF = auto()
    MONITOR = auto()
    BLOCK = auto()
    BLOCK_AT_PERIMETER = auto()
    PERMIT = auto()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config.base_flask_config_rule import (
    BaseFlaskConfigRule,
)
from .httponly_rule import HttpOnlyRuleMixin


class FlaskHttpOnlyRule(HttpOnlyRuleMixin, BaseFlaskConfigRule):
    SETTINGS_VALUE = "SESSION_COOKIE_HTTPONLY"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class LibraryUsage(BaseTsAppMessage):
    def __init__(self, observed_libraries=None):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()

        self.base_url = f"{self.settings.api_url}/agents/v1.0/"
        self.body = {"observations": []}

        if observed_libraries is None:
            return
        for lib in observed_libraries:
            if len(lib.get("used_files")) > 0:
                message = {
                    "id": lib.get("hash_code"),
                    "names": list(lib.get("used_files")),
                }

                self.body["observations"].append(message)

    @property
    def name(self):
        return "library-usage"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "library-usage",
            ]
        )

    @property
    def request_method(self):
        return requests.post

    @fail_loudly("Failed to process LibraryUsage response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Finding, send_finding
from contrast.agent.assess.rules.base_rule import BaseRule


class BaseResponseRule(BaseRule):
    EXCLUDED_RESPONSE_CODES = [301, 302, 307, 404, 410, 500]
    EXCLUDED_CONTENT_TYPES = ["json", "xml"]

    @property
    def name(self):
        raise NotImplementedError

    def is_violated(self, header, body, form_tags, meta_tags):
        raise NotImplementedError

    def is_valid(self, status_code, content_type):
        """
        Rule is valid for analysis if response has matching content-type and status-code
        :return: bool
        """
        return not (
            status_code in self.EXCLUDED_RESPONSE_CODES
            or any(
                [
                    c_type
                    for c_type in self.EXCLUDED_CONTENT_TYPES
                    if c_type in content_type
                ]
            )
        )

    def update_preflight_hash(self, hasher, **kwargs):
        # Response rules do not update the hash
        pass

    def create_finding(self, properties):
        return Finding(self, properties)

    def build_and_append_finding(self, properties, context):
        finding = self.create_finding(properties)
        send_finding(finding, context)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast_vendor.wrapt import register_post_import_hook

from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.settings import Settings
from contrast.utils.patch_utils import wrap_and_watermark

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

MODULE_NAME = "rest_framework.response"
PROPERTY_NAME = "rendered_content"


def build_fget(original_fget):
    def fget_patch(wrapped, instance, args, kwargs):
        del instance

        logger.debug("Hit DRF Response.rendered_content deadzone")
        with scope.contrast_scope():
            return wrapped(*args, **kwargs)

    return wrap_and_watermark(original_fget, fget_patch)


def patch_rest_framework_response(rest_framework_response_module):
    """
    This is a property patch (fget only).
    We can't override fget on the original property object, so we create a new one
    and use the original fset and fdel. If either of these don't exist, they'll
    correctly be set to None on the new property object.
    """
    orig_property = getattr(rest_framework_response_module.Response, PROPERTY_NAME)
    property_with_patch = property(
        fget=build_fget(orig_property.fget),
        fset=orig_property.fset,
        fdel=orig_property.fdel,
    )
    patch_manager.patch(
        rest_framework_response_module.Response, PROPERTY_NAME, property_with_patch
    )


def register_patches():
    """
    This deadzone patch was implemented specifically for a customer, and it's
    off by default.

    Setting agent.python.enable_drf_response_analysis explicitly
    to False will enable this deadzone.
    """
    settings = Settings()
    if settings.config.get("agent.python.enable_drf_response_analysis") is False:
        register_post_import_hook(patch_rest_framework_response, MODULE_NAME)


def reverse_patches():
    rest_framework_response_module = sys.modules.get(MODULE_NAME)
    if not rest_framework_response_module:
        return

    patch_manager.reverse_patches_by_owner(rest_framework_response_module.Response)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators import SUPPORTED_TYPES, BasePropagator
from contrast.utils.assess.tag_utils import merge_tags
from contrast.agent.assess.utils import (
    copy_events,
    get_last_event_ids_from_sources,
    get_properties,
    is_tracked,
)
from contrast.utils.decorators import cached_property
from contrast.utils.string_utils import ensure_string


class JoinPropagator(BasePropagator):
    def _add_tags(self, target_props, source_props, offset):
        for label, tags in source_props.tags.items():
            for tag in tags:
                new_tag = tag.copy_modified(offset)
                target_props.add_existing_tag(label, new_tag)

    def __init__(self, node, preshift, target):
        super().__init__(node, preshift, target)
        self.strings_to_join = self.preshift.args[0] if self.preshift.args else []
        self.separator = self.preshift.obj

    def get_parent_ids(self, *args):
        return get_last_event_ids_from_sources(self.sources)

    @cached_property
    def sources(self):
        return [self.preshift.obj] + self.preshift.args[0]

    @property
    def needs_propagation(self):
        """
        This is a fairly hot code path in some applications; be mindful of performance
        """
        if not isinstance(self.target, SUPPORTED_TYPES):
            return False

        # No need to propagate since the join had no effect
        if len(self.strings_to_join) == 0:
            return False

        if is_tracked(self.target):
            return True

        if not self.preshift:
            return False

        for s in self.sources:
            if is_tracked(s):
                return True

        return False

    def _propagate(self):
        # Offset in the target string
        target_offset = 0

        base_properties = get_properties(self.preshift.obj)

        for i, item in enumerate(self.strings_to_join):
            item_properties = get_properties(item)
            if item_properties is not None:
                self._add_tags(self.target_properties, item_properties, target_offset)
                copy_events(self.target_properties, item_properties)

            # accounts for path-like objects, which have no __len__
            target_offset += len(ensure_string(item))

            if i < len(self.strings_to_join) - 1 and base_properties is not None:
                self._add_tags(self.target_properties, base_properties, target_offset)

            target_offset += len(self.separator)

        if base_properties is not None:
            copy_events(self.target_properties, base_properties)

        merge_tags(self.target_properties.tags)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
greenlet-specific code that pretends to be a `threading.local`.

Fails to import if not running under greenlet.
"""

from __future__ import annotations

from typing import Any
from weakref import WeakKeyDictionary

from greenlet import getcurrent


class GreenThreadLocal:
    """
    threading.local() replacement for greenlets.
    """

    def __init__(self) -> None:
        self.__dict__["_weakdict"] = WeakKeyDictionary()

    def __getattr__(self, name: str) -> Any:
        key = getcurrent()
        try:
            return self._weakdict[key][name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, val: Any) -> None:
        key = getcurrent()
        self._weakdict.setdefault(key, {})[name] = val

    def __delattr__(self, name: str) -> None:
        key = getcurrent()
        try:
            del self._weakdict[key][name]
        except KeyError:
            raise AttributeError(name)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class _Traces(BaseTsAppMessage):
    """This class should only be instantiated by Preflight's response handler"""

    def __init__(self, finding, reportable_request):
        super().__init__()

        self.headers.update({"Report-Hash": finding.preflight.split(",")[-1]})

        self.body = self._build_body(finding, reportable_request)

    @property
    def name(self) -> str:
        return "traces"

    @property
    def path(self) -> str:
        return "traces"

    @property
    def request_method(self) -> str:
        return requests.put

    @fail_loudly("Failed to process Traces response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)

    def _build_body(self, finding, reportable_request):
        base_body = {
            "created": (
                finding.events[-1].timestamp_ms if len(finding.events) > 0 else now_ms()
            ),
            "events": [event.to_json() for event in finding.events],
            "properties": finding.properties,
            "routes": [route.to_json_traces() for route in finding.routes],
            "ruleId": finding.rule_id,
            "session_id": self.settings.config.session_id,
            "tags": self.settings.config.assess_tags,
            "version": finding.version,
        }
        if reportable_request is not None:
            base_body["request"] = reportable_request

        return base_body



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode


string_propagators = [
    CompositeNode(
        {
            "module": "BUILTIN",
            "class_name": "str",
            "policy_patch": False,
            "instance_method": True,
            "source": "OBJ",  # This is the default, some specific nodes may override it
            "target": "RETURN",
        },
        [
            {
                # There is no kwarg corresponding to ARG_0
                "method_name": "concat",
                "source": "OBJ,ARG_0",
                "action": "APPEND",
            },
            {
                # This is string multiplication
                # TODO: PYT-1345 SPLAT makes this prone to FPs
                "method_name": "repeat",
                "action": "SPLAT",
            },
            {
                "method_name": "translate",
                "action": "SPLAT",
            },
            {
                "method_name": [
                    "capitalize",
                    "casefold",
                    "lower",
                    "swapcase",
                    "title",
                    "upper",
                    "encode",
                    "decode",
                ],
                "action": "KEEP",
            },
            {
                "method_name": "ljust",
                "action": "APPEND",
            },
            {
                "method_name": ["zfill", "rjust"],
                "action": "PREPEND",
            },
            {
                "method_name": "center",
                "action": "CENTER",
            },
            {
                # replace() takes no keyword arguments
                "method_name": "replace",
                "source": "OBJ,ARG_1",
                "action": "REPLACE",
            },
            {
                "method_name": "format",
                "source": "OBJ,ALL_ARGS,ALL_KWARGS",
                "action": "FORMAT",
            },
            {
                # cformat uses the % operator, so there are no kwargs
                "method_name": "cformat",
                "source": "OBJ,ALL_ARGS",
                "action": "FORMAT",
            },
            {
                # format_map() takes no keyword arguments
                "method_name": "formatmap",
                "source": "OBJ,ALL_ARGS",
                "action": "FORMAT",
            },
            {
                # fstring formatting is accomplished with string joins, and doesn't use kwargs
                "method_name": "fstring",
                "source": "OBJ,ARG_0",
                "action": "JOIN",
            },
            {
                "method_name": [
                    "strip",
                    "lstrip",
                    "rstrip",
                    "removeprefix",
                    "removesuffix",
                ],
                "action": "REMOVE",
            },
            {
                "method_name": ["subscript", "__getitem__"],
                "action": "SLICE",
            },
            {
                # join() takes no keyword arguments
                "method_name": "join",
                "source": "OBJ,ARG_0",
                "action": "JOIN",
            },
            {
                "method_name": [
                    "split",
                    "splitlines",
                    "rsplit",
                    "partition",
                    "rpartition",
                ],
                "action": "SPLIT",
            },
            {
                # This is str(), bytes(), unicode(), or bytearray().
                # The reported name is slightly misleading
                "instance_method": False,
                "method_name": "CAST",
                "source": "ARG_0,KWARG:object",
                "action": "SPLAT",
            },
            {
                "method_name": "__repr__",
                "action": "REPR",
            },
        ],
    ),
    CompositeNode(
        {
            "module": "operator",
            "target": "RETURN",
            "policy_patch": False,
        },
        [
            {
                "method_name": ["add", "iadd"],
                "source": "ARG_0,ARG_1",
                "action": "APPEND",
            },
            {
                "method_name": "mod",
                "source": "ARG_0,ARG_1",
                "action": "FORMAT",
            },
        ],
    ),
    {
        "module": "BUILTIN",
        "method_name": ["ascii", "format"],
        # No keyword arguments
        "source": "ARG_0",
        "target": "RETURN",
        "action": "REPR",
    },
]


register_propagation_nodes(string_propagators)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Extracted log level data used by both stdlib and native log level filters.
"""

from __future__ import annotations

import asyncio
import contextvars
import logging
import sys

from typing import Any, Callable

from ._base import BoundLoggerBase
from .typing import EventDict, FilteringBoundLogger


# Adapted from the stdlib
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_NAME_TO_LEVEL = {
    "critical": CRITICAL,
    "exception": ERROR,
    "error": ERROR,
    "warn": WARNING,
    "warning": WARNING,
    "info": INFO,
    "debug": DEBUG,
    "notset": NOTSET,
}

_LEVEL_TO_NAME = {
    v: k
    for k, v in _NAME_TO_LEVEL.items()
    if k not in ("warn", "exception", "notset")
}


def add_log_level(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """
    Add the log level to the event dict under the ``level`` key.

    Since that's just the log method name, this processor works with non-stdlib
    logging as well. Therefore it's importable both from `structlog.processors`
    as well as from `structlog.stdlib`.

    .. versionadded:: 15.0.0
    .. versionchanged:: 20.2.0
       Importable from `structlog.processors` (additionally to
       `structlog.stdlib`).
    """
    if method_name == "warn":
        # The stdlib has an alias
        method_name = "warning"

    event_dict["level"] = method_name

    return event_dict


def _nop(self: Any, event: str, *args: Any, **kw: Any) -> Any:
    return None


async def _anop(self: Any, event: str, *args: Any, **kw: Any) -> Any:
    return None


def exception(self: FilteringBoundLogger, event: str, **kw: Any) -> Any:
    kw.setdefault("exc_info", True)

    return self.error(event, **kw)


async def aexception(self: FilteringBoundLogger, event: str, **kw: Any) -> Any:
    # Exception info has to be extracted this early, because it is no longer
    # available once control is passed to the executor.
    if kw.get("exc_info", True) is True:
        kw["exc_info"] = sys.exc_info()

    ctx = contextvars.copy_context()
    return await asyncio.get_running_loop().run_in_executor(
        None,
        lambda: ctx.run(lambda: self.error(event, **kw)),
    )


def make_filtering_bound_logger(min_level: int) -> type[FilteringBoundLogger]:
    """
    Create a new `FilteringBoundLogger` that only logs *min_level* or higher.

    The logger is optimized such that log levels below *min_level* only consist
    of a ``return None``.

    All familiar log methods are present, with async variants of each that are
    prefixed by an ``a``. Therefore, the async version of ``log.info("hello")``
    is ``await log.ainfo("hello")``.

    Additionally it has a ``log(self, level: int, **kw: Any)`` method to mirror
    `logging.Logger.log` and `structlog.stdlib.BoundLogger.log`.

    Compared to using *structlog*'s standard library integration and the
    `structlog.stdlib.filter_by_level` processor:

    - It's faster because once the logger is built at program start; it's a
      static class.
    - For the same reason you can't change the log level once configured. Use
      the dynamic approach of `standard-library` instead, if you need this
      feature.
    - You *can* have (much) more fine-grained filtering by :ref:`writing a
      simple processor <finer-filtering>`.

    :param min_level: The log level as an integer. You can use the constants
        from `logging` like ``logging.INFO`` or pass the values directly. See
        `this table from the logging docs
        <https://docs.python.org/3/library/logging.html#levels>`_ for possible
        values.

    .. versionadded:: 20.2.0
    .. versionchanged:: 21.1.0 The returned loggers are now pickleable.
    .. versionadded:: 20.1.0 The ``log()`` method.
    .. versionadded:: 22.2.0
       Async variants ``alog()``, ``adebug()``, ``ainfo()``, and so forth.
    """

    return _LEVEL_TO_FILTERING_LOGGER[min_level]


def _make_filtering_bound_logger(min_level: int) -> type[FilteringBoundLogger]:
    """
    Create a new `FilteringBoundLogger` that only logs *min_level* or higher.

    The logger is optimized such that log levels below *min_level* only consist
    of a ``return None``.
    """

    def make_method(
        level: int,
    ) -> tuple[Callable[..., Any], Callable[..., Any]]:
        if level < min_level:
            return _nop, _anop

        name = _LEVEL_TO_NAME[level]

        def meth(self: Any, event: str, *args: Any, **kw: Any) -> Any:
            if not args:
                return self._proxy_to_logger(name, event, **kw)

            return self._proxy_to_logger(name, event % args, **kw)

        async def ameth(self: Any, event: str, *args: Any, **kw: Any) -> Any:
            if args:
                event = event % args

            ctx = contextvars.copy_context()
            await asyncio.get_running_loop().run_in_executor(
                None,
                lambda: ctx.run(
                    lambda: self._proxy_to_logger(name, event, **kw)
                ),
            )

        meth.__name__ = name
        ameth.__name__ = f"a{name}"

        return meth, ameth

    def log(self: Any, level: int, event: str, *args: Any, **kw: Any) -> Any:
        if level < min_level:
            return None
        name = _LEVEL_TO_NAME[level]

        if not args:
            return self._proxy_to_logger(name, event, **kw)

        return self._proxy_to_logger(name, event % args, **kw)

    async def alog(
        self: Any, level: int, event: str, *args: Any, **kw: Any
    ) -> Any:
        if level < min_level:
            return None
        name = _LEVEL_TO_NAME[level]
        if args:
            event = event % args

        ctx = contextvars.copy_context()
        return await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: ctx.run(lambda: self._proxy_to_logger(name, event, **kw)),
        )

    meths: dict[str, Callable[..., Any]] = {"log": log, "alog": alog}
    for lvl, name in _LEVEL_TO_NAME.items():
        meths[name], meths[f"a{name}"] = make_method(lvl)

    meths["exception"] = exception
    meths["aexception"] = aexception
    meths["fatal"] = meths["error"]
    meths["afatal"] = meths["aerror"]
    meths["warn"] = meths["warning"]
    meths["awarn"] = meths["awarning"]
    meths["msg"] = meths["info"]
    meths["amsg"] = meths["ainfo"]

    return type(
        "BoundLoggerFilteringAt%s"
        % (_LEVEL_TO_NAME.get(min_level, "Notset").capitalize()),
        (BoundLoggerBase,),
        meths,
    )


# Pre-create all possible filters to make them pickleable.
BoundLoggerFilteringAtNotset = _make_filtering_bound_logger(NOTSET)
BoundLoggerFilteringAtDebug = _make_filtering_bound_logger(DEBUG)
BoundLoggerFilteringAtInfo = _make_filtering_bound_logger(INFO)
BoundLoggerFilteringAtWarning = _make_filtering_bound_logger(WARNING)
BoundLoggerFilteringAtError = _make_filtering_bound_logger(ERROR)
BoundLoggerFilteringAtCritical = _make_filtering_bound_logger(CRITICAL)

_LEVEL_TO_FILTERING_LOGGER = {
    CRITICAL: BoundLoggerFilteringAtCritical,
    ERROR: BoundLoggerFilteringAtError,
    WARNING: BoundLoggerFilteringAtWarning,
    INFO: BoundLoggerFilteringAtInfo,
    DEBUG: BoundLoggerFilteringAtDebug,
    NOTSET: BoundLoggerFilteringAtNotset,
}



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_VALIDATED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "BASE64_ENCODED",
    "CSS_ENCODED",
    "CSV_ENCODED",
    "HTML_ENCODED",
    "JAVASCRIPT_ENCODED",
    "JAVA_ENCODED",
    "LDAP_ENCODED",
    "OS_ENCODED",
    "SQL_ENCODED",
    "URL_ENCODED",
    "VBSCRIPT_ENCODED",
    "XML_ENCODED",
    "XPATH_ENCODED",
]


trust_boundary_violation_triggers = [
    {
        "module": "quart.sessions",
        "class_name": "SecureCookieSession",
        "method_name": "__setitem__",
        # no keyword arguments
        "source": "ARG_1",
    },
    {
        "module": "quart.sessions",
        "class_name": "SecureCookieSession",
        "method_name": "setdefault",
        # no keyword arguments
        "source": "ARG_1",
    },
    {
        # Used by bottle for sessions. No keyword arguments
        "module": "bottle_session",
        "class_name": "Session",
        "method_name": "__setitem__",
        "source": "ARG_1",
    },
    {
        # WSGI pluggable for sessions. No keyword arguments
        "module": "beaker.session",
        "class_name": "Session",
        "method_name": "__setitem__",
        "source": "ARG_1",
    },
    {
        # WSGI pluggable for sessions. No keyword arguments
        "module": "beaker.session",
        "class_name": "Session",
        "method_name": "setdefault",
        "source": "ARG_1",
    },
    {
        "module": "django.contrib.sessions.backends.base",
        "class_name": "SessionBase",
        "method_name": ["__setitem__", "setdefault"],
        # no keyword arguments
        "source": "ARG_1",
    },
    {
        "module": "flask.sessions",
        "class_name": "SecureCookieSession",
        "method_name": ["__setitem__", "setdefault"],
        # no keyword arguments
        "source": "ARG_1",
    },
    {
        "module": "pyramid.session",
        "class_name": "CookieSession",
        "method_name": ["__setitem__", "setdefault"],
        # no keyword arguments
        "source": "ARG_1",
        "policy_patch": False,
    },
    {
        # No kwargs. This is a fake node used for triggering trust-boundary-
        # -violation for proxied dictionaries returned by
        # starlette.requests.Request.session. class_name is sort of nonsense
        # here. see starlette_patches.py
        "module": "starlette.sessions",
        "class_name": "dict",
        "method_name": "__setitem__",
        "source": "ARG_1",
        "policy_patch": False,
    },
    {
        # Used by aiohttp for sessions. No keyword arguments
        "module": "aiohttp_session",
        "class_name": "Session",
        "method_name": "__setitem__",
        "source": "ARG_1",
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "trust-boundary-violation",
        trust_boundary_violation_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.utils.decorators import fail_quietly
from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    build_args_from_function,
    build_key,
    get_normalized_uri,
)
from pyramid.interfaces import IView, IViewClassifier, IRouteRequest, IRoutesMapper
from zope.interface import Interface

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def get_iface_for_route(registry, pyramid_route):
    return registry.queryUtility(IRouteRequest, name=pyramid_route.name)


def _get_view_func(registry, pyramid_route):
    request_iface = get_iface_for_route(registry, pyramid_route)

    if request_iface is None:
        view_func = None
    else:
        view_func = registry.adapters.lookup(
            (IViewClassifier, request_iface, Interface), IView, name="", default=None
        )
    return view_func


def create_pyramid_routes(registry):
    """
    Returns all the routes registered to an app as a dict

    The Pyramid Request object holds a reference to the url route registry

    :param registry: Pyramid Registry
    :return: Returns a dict of key = id, value = api.Route.
    """
    routes = {}
    mapper = registry.queryUtility(IRoutesMapper)

    if mapper is None:
        return routes

    for pyramid_route in mapper.get_routes():
        view_func = _get_view_func(registry, pyramid_route)

        if view_func is not None:
            route = build_pyramid_route(pyramid_route.name, view_func)
            route_id = str(id(view_func))

            methods = pyramid_route.predicates or DEFAULT_ROUTE_METHODS

            for method_type in methods:
                key = build_key(route_id, method_type)

                url = get_normalized_uri(pyramid_route.path)
                routes[key] = Route(verb=method_type, url=url, route=route)
        else:
            logger.debug("Unable to add %s to route coverage.", pyramid_route.path)

    return routes


def build_pyramid_route(pyramid_route_name, view_func):
    method_arg_names = build_args_from_function(view_func)

    return pyramid_route_name + method_arg_names


@fail_quietly("Unable to get route coverage", return_value={})
def get_route_coverage(registry):
    return create_pyramid_routes(registry)


@fail_quietly("Unable to find new routes")
def get_view_func(request_path, routes_list, registry):
    if not request_path:
        logger.debug("No path info for pyramid request")
        return None

    # Ideally we would like to call get_route but
    # there is no direct relationship between the wsgi
    # request path and the name of the route so we must
    # iterate over all the routes to find it by the path
    # pyramid_route = mapper.get_route(name)

    matching_routes = [x for x in routes_list if x.path == request_path]
    if not matching_routes:
        return None

    return _get_view_func(registry, matching_routes[0])


@fail_quietly("Unable to build route", return_value="")
def build_route(view_func, url):
    route_name = url.replace("/", "")
    return build_pyramid_route(route_name, view_func)



"""
Represents the Cache-Control header
"""
import re

class UpdateDict(dict):
    """
    Dict that has a callback on all updates
    """
    # these are declared as class attributes so that
    # we don't need to override constructor just to
    # set some defaults
    updated = None
    updated_args = None

    def _updated(self):
        """
        Assign to new_dict.updated to track updates
        """
        updated = self.updated
        if updated is not None:
            args = self.updated_args
            if args is None:
                args = (self,)
            updated(*args)

    def __setitem__(self, key, item):
        dict.__setitem__(self, key, item)
        self._updated()

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self._updated()

    def clear(self):
        dict.clear(self)
        self._updated()

    def update(self, *args, **kw):
        dict.update(self, *args, **kw)
        self._updated()

    def setdefault(self, key, value=None):
        val = dict.setdefault(self, key, value)
        if val is value:
            self._updated()
        return val

    def pop(self, *args):
        v = dict.pop(self, *args)
        self._updated()
        return v

    def popitem(self):
        v = dict.popitem(self)
        self._updated()
        return v


token_re = re.compile(
    r'([a-zA-Z][a-zA-Z_-]*)\s*(?:=(?:"([^"]*)"|([^ \t",;]*)))?')
need_quote_re = re.compile(r'[^a-zA-Z0-9._-]')


class exists_property(object):
    """
    Represents a property that either is listed in the Cache-Control
    header, or is not listed (has no value)
    """
    def __init__(self, prop, type=None):
        self.prop = prop
        self.type = type

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return self.prop in obj.properties

    def __set__(self, obj, value):
        if (self.type is not None
            and self.type != obj.type):
            raise AttributeError(
                "The property %s only applies to %s Cache-Control" % (
                    self.prop, self.type))

        if value:
            obj.properties[self.prop] = None
        else:
            if self.prop in obj.properties:
                del obj.properties[self.prop]

    def __delete__(self, obj):
        self.__set__(obj, False)


class value_property(object):
    """
    Represents a property that has a value in the Cache-Control header.

    When no value is actually given, the value of self.none is returned.
    """
    def __init__(self, prop, default=None, none=None, type=None):
        self.prop = prop
        self.default = default
        self.none = none
        self.type = type

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        if self.prop in obj.properties:
            value = obj.properties[self.prop]
            if value is None:
                return self.none
            else:
                return value
        else:
            return self.default

    def __set__(self, obj, value):
        if (self.type is not None
            and self.type != obj.type):
            raise AttributeError(
                "The property %s only applies to %s Cache-Control" % (
                    self.prop, self.type))
        if value == self.default:
            if self.prop in obj.properties:
                del obj.properties[self.prop]
        elif value is True:
            obj.properties[self.prop] = None # Empty value, but present
        else:
            obj.properties[self.prop] = value

    def __delete__(self, obj):
        if self.prop in obj.properties:
            del obj.properties[self.prop]


class CacheControl(object):

    """
    Represents the Cache-Control header.

    By giving a type of ``'request'`` or ``'response'`` you can
    control what attributes are allowed (some Cache-Control values
    only apply to requests or responses).
    """

    update_dict = UpdateDict

    def __init__(self, properties, type):
        self.properties = properties
        self.type = type

    @classmethod
    def parse(cls, header, updates_to=None, type=None):
        """
        Parse the header, returning a CacheControl object.

        The object is bound to the request or response object
        ``updates_to``, if that is given.
        """
        if updates_to:
            props = cls.update_dict()
            props.updated = updates_to
        else:
            props = {}
        for match in token_re.finditer(header):
            name = match.group(1)
            value = match.group(2) or match.group(3) or None
            if value:
                try:
                    value = int(value)
                except ValueError:
                    pass
            props[name] = value
        obj = cls(props, type=type)
        if updates_to:
            props.updated_args = (obj,)
        return obj

    def __repr__(self):
        return '<CacheControl %r>' % str(self)

    # Request values:
    # no-cache shared (below)
    # no-store shared (below)
    # max-age shared  (below)
    max_stale = value_property('max-stale', none='*', type='request')
    min_fresh = value_property('min-fresh', type='request')
    # no-transform shared (below)
    only_if_cached = exists_property('only-if-cached', type='request')

    # Response values:
    public = exists_property('public', type='response')
    private = value_property('private', none='*', type='response')
    no_cache = value_property('no-cache', none='*')
    no_store = exists_property('no-store')
    no_transform = exists_property('no-transform')
    must_revalidate = exists_property('must-revalidate', type='response')
    proxy_revalidate = exists_property('proxy-revalidate', type='response')
    max_age = value_property('max-age', none=-1)
    s_maxage = value_property('s-maxage', type='response')
    s_max_age = s_maxage
    stale_while_revalidate = value_property(
        'stale-while-revalidate', type='response')
    stale_if_error = value_property('stale-if-error', type='response')

    def __str__(self):
        return serialize_cache_control(self.properties)

    def copy(self):
        """
        Returns a copy of this object.
        """
        return self.__class__(self.properties.copy(), type=self.type)


def serialize_cache_control(properties):
    if isinstance(properties, CacheControl):
        properties = properties.properties
    parts = []
    for name, value in sorted(properties.items()):
        if value is None:
            parts.append(name)
            continue
        value = str(value)
        if need_quote_re.search(value):
            value = '"%s"' % value
        parts.append('%s=%s' % (name, value))
    return ', '.join(parts)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule
from contrast.agent.policy.utils import CompositeNode


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_SSRF",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_SSRF",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
]

ssrf_triggers = [
    CompositeNode(
        {
            "action": "SSRF",
        },
        [
            {
                "module": "urllib.request",
                "method_name": "urlopen",
                "source": "ARG_0,KWARG:url",
                "protect_mode": True,
            },
            {
                # Serves to trigger both sync and async http.get, http.post, etc
                "module": "httpx._client",
                "class_name": "BaseClient",
                "method_name": "build_request",
                "source": "ARG_1,KWARG:url",
                "protect_mode": True,
            },
            {
                "module": "aiohttp",
                "class_name": "ClientSession",
                "method_name": "_request",
                # ARG_0 is HTTP method. They're all defined in their file hdrs.py.
                # ARG_1 is the url. Could be both string and 'URL' class which comes from a yarl module
                "source": "ARG_0,ARG_1,KWARG:method,KWARG:str_or_url",
                "protect_mode": True,
            },
        ],
    ),
    {
        "module": "http.client",
        "class_name": "HTTPConnection",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:host",
        "action": "DEFAULT",
    },
    {
        # This also serves as a trigger for the request method
        "module": "http.client",
        "class_name": "HTTPConnection",
        "method_name": "putrequest",
        # ARG_0 is HTTP method, so no regex is used to evaluate
        # ARG_1 is a path + querystring only, which isn't vulnerable to SSRF
        "source": "ARG_0,KWARG:method",
        "action": "DEFAULT",
        "protect_mode": True,
    },
    {
        # This is the underlying call for functions like requests.get, etc.
        # This function calls Session.request under the hood, so it's
        # technically duplicated by the Session.request node below, but it
        # might lead to better storytelling to keep it.
        "module": "requests.api",
        "method_name": "request",
        "source": "ARG_1,KWARG:url",
        "action": "SSRF",
    },
    {
        # This is the underlying call for methods like Session.get, etc.
        "module": "requests.sessions",
        "class_name": "Session",
        "method_name": "request",
        "source": "ARG_1,KWARG:url",
        "action": "SSRF",
    },
    {
        "module": "requests.models",
        "class_name": "Request",
        "method_name": "__init__",
        "source": "ARG_1,KWARG:url",
        "action": "SSRF",
    },
]

register_trigger_rule(
    DataflowRule.from_nodes(
        "ssrf",
        ssrf_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import copy_from


class ReductivePropagator(BasePropagator):
    def _propagate(self):
        copy_from(self.target, self.first_source, 0, self.node.untags)

        source_index = 0
        target_index = 0

        remove_ranges = []
        current_range = None

        # loop over the target, the result of the delete
        # every range of characters that it differs from the source
        # represents a section that was deleted. these sections
        # need to have their tags updated
        while target_index < len(self.target) and source_index < len(self.first_source):
            target_char = self.target[target_index]
            source_char = self.first_source[source_index]

            if target_char == source_char:
                target_index += 1

                if current_range:
                    current_range.stop = source_index

                    remove_ranges.append(current_range)
                    current_range = None
            elif current_range is None:
                current_range = AdjustedSpan(target_index, len(self.first_source))

            source_index += 1

        if source_index != len(self.first_source):
            remove_ranges.append(AdjustedSpan(target_index, len(self.first_source)))

        self.target_properties.delete_tags_at_ranges(remove_ranges)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Ssrf(BaseRule):
    """
    Ssrf Protection rule
    Currently in BETA.
    """

    RULE_NAME = "ssrf"

    def is_postfilter(self):
        return False

    def build_sample(self, evaluation, url, **kwargs):
        sample = self.build_base_sample(evaluation)
        if url is not None:
            sample.details["url"] = url
        return sample



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Instrumentation for exec/eval.

This was defined outside of policy.json because we need to pass globals/locals from the
frame in which function was originally called
"""
import ast
import copy
from sys import _getframe as getframe

from contrast_vendor.wrapt import register_post_import_hook

import builtins
import contrast
from contrast.agent import scope
from contrast.agent.policy import registry
from contrast.agent.assess.policy import trigger_policy
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.applies.assess.unsafe_code_execution import (
    apply_rule as apply_unsafe_code_exec_rule,
)
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import wrap_and_watermark
from contrast_rewriter import PropagationRewriter, populate_operator_module


INSTRUMENTED_FRAME_DEPTH = 1

orig_compile = builtins.compile


@fail_quietly("Error applying rule for exec/eval patch")
def apply_rule(rule_applicator, orig_func, result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if not (context and context.propagate_assess):
        return

    with scope.contrast_scope():
        rule_applicator("BUILTIN", orig_func.__name__, result, args, kwargs)


@fail_quietly("Error running path traversal assess rule")
def apply_pt_rule(module, method, result, args, kwargs):
    trigger_rule = registry.get_triggers_by_rule("path-traversal")

    trigger_nodes = trigger_rule.find_trigger_nodes(module, method)

    trigger_policy.apply(trigger_rule, trigger_nodes, result, args, kwargs)


def rewrite(mode, code):
    rewriter = PropagationRewriter()
    return ast.fix_missing_locations(rewriter.visit(ast.parse(code, mode=mode)))


@fail_quietly("Error when applying rewriter to eval/exec/compiled input")
def apply_rewriter(mode, code):
    if not isinstance(code, str):
        return None

    with scope.contrast_scope():
        return orig_compile(
            rewrite(mode, code),
            filename="<internal>",
            mode=mode,
        )


def build_exec_eval_patch(orig_func, _, rule_applicator, mode):
    def exec_eval_patch(code, globs=None, locs=None):
        """
        Run exec/eval call with proper context to adjust for current frame

        Code ported from six module
        See https://github.com/benjaminp/six/blob/master/six.py#L694

        Reapplying the context from the 3rd frame (from top of stack) is necessary
        because the globals and locals in that frame are used in the original call to
        exec/eval. The exception to this is if the caller passes custom globals/locals
        to the function.

        If we fail provide this context we will see a number of errors due to things
        not defined in the scope of this function upon calling the original function
        definition.
        """
        result = None

        if globs is None:
            frame = getframe(INSTRUMENTED_FRAME_DEPTH)

            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs

        # Ensure our rewriter patches are in (global) scope for the code about to be executed
        populate_operator_module(globs)

        try:
            code_to_run = apply_rewriter(mode, code) or code
            result = orig_func(code_to_run, globs, locs)
        except Exception:
            result = None
            raise
        finally:
            apply_rule(rule_applicator, orig_func, result, (code,), {})

        return result

    # NOTE: do not wrap since it will potentially affect stack level
    return exec_eval_patch


# NOTE: maybe this should be moved to a utility module
def find_arg(args, kwargs, idx, kw=None, default=None, pop=False):
    if kw and kw in kwargs:
        return kwargs.pop(kw) if pop else kwargs[kw]

    if len(args) <= idx:
        return default

    return args[idx]


def build_compile_patch(orig_func, _, rule_applicator):
    def compile_patch(wrapped, instance, args, kwargs):
        del instance

        if scope.in_contrast_scope():
            return wrapped(*args, **kwargs)

        result = None
        orig_kwargs = copy.copy(kwargs)

        try:
            code = find_arg(args, kwargs, 0, kw="source", pop=True)
            mode = find_arg(args, kwargs, 2, kw="mode", default="exec")

            with scope.contrast_scope():
                code_to_compile = rewrite(mode, code)
                result = orig_compile(code_to_compile, *args[1:], **kwargs)
        except Exception:
            result = None
            raise
        finally:
            apply_rule(rule_applicator, orig_func, result, args, orig_kwargs)

        return result

    return wrap_and_watermark(orig_func, compile_patch)


def patch_exec_and_eval(builtins_module):
    build_and_apply_patch(
        builtins_module,
        "eval",
        build_exec_eval_patch,
        (apply_unsafe_code_exec_rule, "eval"),
    )

    build_and_apply_patch(
        builtins_module,
        "exec",
        build_exec_eval_patch,
        (apply_unsafe_code_exec_rule, "exec"),
    )

    build_and_apply_patch(
        builtins_module, "compile", build_compile_patch, (apply_unsafe_code_exec_rule,)
    )


def register_patches():
    register_post_import_hook(patch_exec_and_eval, builtins.__name__)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import logging

import contrast
from contrast.assess_extensions import cs_str
from contrast.utils.loggers.structlog import init_structlog
from contrast.utils.configuration_utils import get_hostname
from contrast.utils.namespace import Namespace
from contrast.configuration import ConfigOptionSrc, CONTRAST_UI_SRC

from contrast_vendor import structlog

from . import (
    DEFAULT_LOG_LEVEL,
    DEFAULT_LOG_PATH,
    DEFAULT_PROGNAME,
    LOGGER_NAME,
    SECURITY_LOGGER_NAME,
)

STDOUT = "STDOUT"
STDERR = "STDERR"


class module(Namespace):
    initialized: bool = False


def setup_basic_agent_logger(level=logging.INFO):
    """
    Setup a logger without any user-supplied configuration, with defaults:
        1. log to stdout
        2. log in INFO level
        3. with default progname

    The logger created here is expected to be overridden with config values
    provided later on in the middleware creation cycle.

    Note regarding init lock:
        Previously we discussed having a lock as part of logger init but since setup_basic_agent_logger is only
        called when the module is loaded (i.e during module import + executed) it should be safe not to explicitly
        have one due to the locking mechanisms used in the import machinery.
    """
    if not module.initialized:
        logger = logging.getLogger(LOGGER_NAME)
        logger.addHandler(logging.StreamHandler(sys.stdout))
        _set_handler(logger, "STDOUT", DEFAULT_PROGNAME)
        logger.setLevel(level)

        init_structlog()
        module.initialized = True

    return structlog.getLogger(LOGGER_NAME)


def setup_agent_logger(config):
    """
    Initialize the agent logger with configurations.
    :param config: instance of AgentConfig or dict
    :return: None
    """
    config = config if config else {}
    # the or handles the case of empty string
    path = config.get("agent.logger.path", DEFAULT_LOG_PATH) or DEFAULT_LOG_PATH
    level = (
        config.get("agent.logger.level", DEFAULT_LOG_LEVEL).upper() or DEFAULT_LOG_LEVEL
    )

    logger = logging.getLogger(LOGGER_NAME)

    _set_logger_info(logger, config, path, level)
    _lock_logger(logger, config)

    cs_str.initialize_logger(structlog.getLogger(LOGGER_NAME))


def setup_security_logger(cfg):
    config = cfg if cfg else {}

    if not config.get("protect.enable", None):
        return

    path = config.get("agent.security_logger.path")
    level = config.get("agent.security_logger.level").upper()

    logger = logging.getLogger(SECURITY_LOGGER_NAME)

    _set_level(logger, level)

    handler = _get_handler(path)
    logger.addHandler(handler)

    fmt = (
        f"%(asctime)s {get_hostname()} CEF:0|Contrast Security|Contrast Agent Python|{contrast.__version__}|"
        "SECURITY|%(message)s|%(level)s|'pri=%(rule_id)s' 'src=%(source_ip)s' 'spt=%(source_port)s' "
        "'request=%(request_url)s' 'requestMethod=%(request_method)s' 'app=%(application)s' "
        "'outcome=%(outcome)s'"
    )

    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)


def security_log_msg(msg, app_name, rule_name, outcome):
    context = contrast.CS__CONTEXT_TRACKER.current()
    ip = port = url = method_name = "-"

    logger = logging.getLogger(SECURITY_LOGGER_NAME)

    if context is not None:
        ip = context.request.client_addr or "-"
        port = context.request.host_port
        method_name = context.request.method
        url = context.request.path

    log_context = dict(
        level=logger.level,
        rule_id=rule_name,
        source_ip=ip,
        source_port=port,
        request_url=url,
        request_method=method_name,
        application=app_name,
        outcome=outcome,
    )

    logger.warning(msg, extra=log_context)


def reset_agent_logger(log_path, log_level, config):
    """
    Reset agent logger path and/or level after the logger has already been created.

    Note that this function will return early if the lock_ attributes are not set,
    as this is what we set when we initialize the logger.

    Also note that progname is never reset so we use the one already set to the logger.

    :return: Bool if any logger value is reset
    """
    logger = logging.getLogger(LOGGER_NAME)

    if not hasattr(logger, "lock_path") or not hasattr(logger, "lock_level"):
        structlog.getLogger(LOGGER_NAME).debug(
            "Will not reset agent logger without set lock_attr values."
        )
        return False

    is_reset = False

    if not logger.lock_path and log_path:
        current_handler = logger.handlers[0]
        progname = current_handler.filters[0].progname
        _set_handler(logger, log_path, progname)
        # print so it shows up in STDOUT
        print(f"Contrast Agent Logger updated path to {log_path}")
        is_reset = True

        if config is not None:
            config.put(
                "agent.logger.path",
                log_path,
                src=ConfigOptionSrc(
                    name="contrast.agent.logger.path", source=CONTRAST_UI_SRC
                ),
            )

    if not logger.lock_level and log_level:
        # Avoid circular import
        from contrast.agent.agent_lib import update_log_options

        _set_level(logger, log_level)

        # print so it shows up in STDOUT
        print(f"Contrast Agent Logger updated level to {log_level}")

        if update_log_options(log_level):
            print(f"Contrast Agent Lib Logger updated level to {log_level}")

        is_reset = True

        if config is not None:
            config.put(
                "agent.logger.level",
                log_level,
                src=ConfigOptionSrc(
                    name="contrast.agent.logger.level", source=CONTRAST_UI_SRC
                ),
            )

    return is_reset


def _set_logger_info(logger, config, path, level):
    progname = config.get("agent.logger.progname", DEFAULT_PROGNAME)

    _set_handler(logger, path, progname)
    _set_level(logger, level)


def _set_handler(logger, path, progname):
    """
    A logger's handler is what determines where the log records will be printed to
    and what format they will have.

    To reset a handler, we delete the existing handlers and create a new one.

    CONTRAST-39746 defined the datetime format as ISO_8601. The one here is
    without ms as the logger doesn't natively support both ms and time zone at this time.
    """
    handler = _get_handler(path)
    program_filter = AgentFilter(progname=progname)
    handler.addFilter(program_filter)

    # empty all handlers so there is only one logging handler with this config
    logger.handlers = []
    logger.addHandler(handler)


def _get_handler(path):
    if path == STDOUT:
        handler = logging.StreamHandler(sys.stdout)
    elif path == STDERR:
        handler = logging.StreamHandler(sys.stderr)
    else:
        try:
            handler = logging.FileHandler(path)
        except Exception as e:
            print(e)
            # path could be '' or None
            handler = logging.StreamHandler()

    return handler


def _set_level(logger, level: str) -> None:
    if level.upper() == "TRACE":
        level = "DEBUG"
        print("Contrast Python Agent: TRACE logging is equivalent to DEBUG")
    try:
        logger.setLevel(level)
    except ValueError:
        # this fails validation if the level is an invalid value
        logger.setLevel(DEFAULT_LOG_LEVEL)


def _lock_logger(logger, config):
    """
    Determine if to lock the logger by looking at if the config had specified values
    for path and level.

    Locking the logger means its path and/or level cannot be changed later on by
    TS server feature values.
    """
    _lock_logger_attr(logger, "path", config)
    _lock_logger_attr(logger, "level", config)


def _lock_logger_attr(logger, logger_attr, config):
    """
    Assign a lock_{logger_attr} attribute to the logger.
    logger.logger_attr will be True if the user provided a legitimate string,
    but will be False if the attr value is '' or None.
    """
    attr_value = config.get(f"agent.logger.{logger_attr}")
    setattr(logger, f"lock_{logger_attr}", bool(attr_value))


class AgentFilter(logging.Filter):
    def __init__(self, progname=None):
        self.progname = progname
        super().__init__()

    def filter(self, record):
        record.progname = self.progname
        return super().filter(record)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

import contrast

from aiohttp.web import StreamResponse
from aiohttp.web_urldispatcher import DynamicResource

from contrast.aiohttp import sources
from contrast.agent import scope
from contrast.agent.middlewares.route_coverage.aiohttp_routes import (
    create_aiohttp_routes,
)
from contrast.agent.middlewares.route_coverage.common import build_route
from contrast.agent.policy.trigger_node import TriggerNode
from contrast_vendor import structlog as logging
from contrast.utils.decorators import cached_property
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.middlewares.response_wrappers.aiohttp_response_wrapper import (
    AioHttpResponseWrapper,
)

from contrast.utils.decorators import fail_quietly
from contrast.utils import Profiler

logger = logging.getLogger("contrast")


class AioHttpMiddleware(BaseMiddleware):
    __middleware_version__ = 1  # Aiohttp new-style middleware

    # Since there is no way to get the `app` instance, on startup of AioHttp,
    # until the first request, hence we will not have `app` finder logic.
    @scope.with_contrast_scope
    def __init__(self, app_name: Optional[str] = None) -> None:
        self.app = None
        self.app_name = app_name or "aiohttp"
        super().__init__()

    async def __call__(self, request, handler) -> StreamResponse:
        if contrast.CS__CONTEXT_TRACKER.current():
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request context")
            return await handler(request)

        self.app = request.app

        self.request_path = request.path

        with Profiler(self.request_path):
            with scope.contrast_scope():
                environ = await sources.aiohttp_request_to_environ(request)

            context = self.should_analyze_request(environ)
            if context:
                with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                    return await self.call_with_agent(context, request, handler)

            return await self.call_without_agent_async(request, handler)

    async def call_with_agent(self, context, request, handler) -> StreamResponse:
        self.log_start_request_analysis()

        sources.track_aiohttp_request_sources(context, request)

        try:
            self.prefilter()

            response = await handler(request)

            with scope.contrast_scope():
                wrapped_response = AioHttpResponseWrapper(response)

            context.extract_response_to_context(wrapped_response)

            self.postfilter(context)
            self.check_for_blocked(context)

            return response

        finally:
            self.handle_ensure(context, request)
            self.log_end_request_analysis()
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, request, handler) -> StreamResponse:
        super().call_without_agent()
        with scope.contrast_scope():
            return await handler(request)

    @fail_quietly("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_aiohttp_routes(self.app)

    @fail_quietly("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_route(url, view_func)

    @fail_quietly("Unable to get view func")
    def get_view_func(self, request):
        if not self.request_path:
            return None

        view_func = None

        # This approach has at worst O(_resources) performance
        # but it's a first attempt at implementing a sync
        # version of aiohttp.web_urldispatches.UrlDispatcher.resolve
        for app_route in self.app.router._resources:
            _app_route = (
                app_route._formatter
                if isinstance(app_route, DynamicResource)
                else app_route._path
            )
            if _app_route == self.request_path:
                routes = app_route._routes
                for route in routes:
                    if route.method == request.method:
                        return route.handler

        return view_func

    @cached_property
    def trigger_node(self):
        """
        Used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "aiohttp"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from datetime import timedelta


class SessionAgeRuleMixin(object):
    @property
    def name(self):
        return "session-timeout"

    def is_violated(self, value):
        """
        A value of 30 mins or less is considered safe

        Flask represents this value as either a timedelta or as an integer in seconds.
        All other frameworks represent it as an integer representing seconds.
        """
        if self.count_threshold_reached():
            return False
        if isinstance(value, timedelta):
            return value > timedelta(minutes=30)

        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                pass

        return value is None or value > 30 * 60



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contextvars

from contextlib import contextmanager


class CvarContextTracker(object):
    """
    This descriptor class makes it simpler to set/get a ContextVar.
    Classes instantiating this descriptor must pass the name of the instance variable to which
    the ContextVar is assigned to.

    For example, a class that uses this descriptor as such:

    class SomeClass:
        my_desc = CvarContextTracker("sample_attr")

        # must have an instance variable defined as such:
        def __init__(self):
            self.sample_attr = ContextVar(...)
    """

    def __init__(self, name):
        self._cvar_instance_name = name

    def __set__(self, instance, value):
        getattr(instance, self._cvar_instance_name).set(value)

    def __get__(self, instance, instance_type=None):
        val = None

        try:
            val = getattr(instance, self._cvar_instance_name).get()
        except LookupError:
            pass

        return val


class ContextTracker(object):
    CURRENT_CONTEXT = "CURRENT_CONTEXT"
    _CVAR_INSTANCE_NAME = "_cvar"
    _CVAR_VAR_NAME = "contrast_request_context"
    _DEFAULT_CVAR_VALUE = None

    request_context = CvarContextTracker(_CVAR_INSTANCE_NAME)

    def __init__(self):
        self._cvar = contextvars.ContextVar(
            self._CVAR_VAR_NAME, default=self._DEFAULT_CVAR_VALUE
        )

    def get(self):
        return self.request_context

    def set(self, value):
        self.request_context = value

    def delete(self):
        self.request_context = None

    def set_current(self, value):
        self.request_context = value

    def delete_current(self):
        self.request_context = None

    @contextmanager
    def lifespan(self, context):
        self.set_current(context)
        try:
            yield context
        finally:
            self.delete_current()

    def current(self):
        return self.request_context



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import traceback
from functools import lru_cache

from contrast import (
    AGENT_CURR_WORKING_DIR,
    SORTED_SYS_PATH,
)
from contrast.api.trace_event import TraceStack
from contrast.utils.decorators import fail_quietly
from contrast.utils.library_reader.library_reader import (
    get_active_library_names_from_pkg,
)
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


APPLIES_MARKER = "cs__"
PATCH_MARKER = "__cs"
PY_FILE_EXTENSION = ".py"
UTILS_MODULES = "contrast/utils"
NONETYPE = "NoneType"

CONTRAST_EXTENSIONS = (
    "contrast_vendor",
    "contrast_rewriter",
    "contrast/assess_extensions",
    "contrast/patches",
)
DJANGO_EXCEPTION_PATH = "core/handlers/exception.py"
DJANGO_DEPRECATION_PATH = "utils/deprecation.py"
SITE_PACKAGES = "site-packages"


@fail_quietly("Failed to build stacktrace for event", return_value=[])
def build_stack():
    # PERF: only grab 30 elements to not use as much memory.
    # But the clean step will likely remove more elements.
    return traceback.extract_stack(limit=30)


@fail_quietly("Failed to clean stacktrace for event", return_value=[])
def clean_stack(frames, depth=20):
    frames.reverse()  # moves most recent to the front

    frames = [frame for frame in frames if acceptable_frame(frame)]

    return to_element_list(frames[:depth])


def acceptable_frame(frame):
    """
    Return true if frame does NOT contain contrast or __cs
    """
    ignore_string = "/contrast/"

    if isinstance(frame, tuple):
        return (
            ignore_string not in frame[0]
            and UTILS_MODULES not in frame[0]
            and not any(extension in frame[0] for extension in CONTRAST_EXTENSIONS)
            and not frame[2].startswith(APPLIES_MARKER)
            and not frame[2].startswith(PATCH_MARKER)
            and not frame[0].endswith(DJANGO_EXCEPTION_PATH)
            and not frame[0].endswith(DJANGO_DEPRECATION_PATH)
        )

    return (
        ignore_string not in frame.filename
        and UTILS_MODULES not in frame.filename
        and not any(extension in frame.filename for extension in CONTRAST_EXTENSIONS)
        and not frame.name.startswith(APPLIES_MARKER)
        and not frame.name.startswith(PATCH_MARKER)
        and not frame.filename.endswith(DJANGO_EXCEPTION_PATH)
        and not frame.filename.endswith(DJANGO_DEPRECATION_PATH)
    )


def build_and_clean_stack(depth=10):
    """
    Perform both build and clean steps.
    """
    frames = build_stack()
    return clean_stack(frames, depth=depth)


def to_element_list(frames):
    return [y for y in [to_element(x) for x in frames] if y is not None]


@fail_quietly("Failed to convert event")
def to_element(summary):
    if not summary:
        return None

    if isinstance(summary, tuple):
        # in python 2 traceback returns a tuple
        path = summary[0]
        method = summary[2]
        line_number = summary[1]
    else:
        # python 3 is a FrameSummary
        path = summary.filename
        method = summary.name
        line_number = summary.lineno

    element = TraceStack()
    element.line_number = line_number

    element.file_name = filename_formatter(path)
    element.declaring_class = element.file_name
    element.method_name = method

    return element


@fail_quietly("Unable to create file_name")
@lru_cache(maxsize=512)
def filename_formatter(file_name):
    # PERF: This method is called hundreds of times, so be mindful
    # of what additional computations are added. Small changes will have
    # significant performance impacts.

    if file_name.startswith("<frozen"):
        return file_name

    formatted = None

    if file_name.startswith(AGENT_CURR_WORKING_DIR):
        formatted = file_name.replace(AGENT_CURR_WORKING_DIR, "")
    else:
        for sys_path in SORTED_SYS_PATH:
            if file_name.startswith(sys_path):
                formatted = file_name.replace(sys_path, "")
                break

    formatted = formatted or file_name

    return formatted.replace("/", ".").lstrip(".")


def in_custom_code():
    current_stack_trace = build_and_clean_stack(depth=4)
    for trace in current_stack_trace:
        if trace:
            file_name = trace.file_name
            if file_name.endswith(".py") or file_name.endswith(".pyc"):
                file_name = file_name[:-3]
            if _is_custom_module(file_name):
                return True
    return False


def _is_custom_module(file_name):
    """

    Given a file name, determines if the file is a custom module.

    If a library name is in the file_name, we can say that the file is part of that library
    Obviously this can go wrong if users are naming their custom modules library names but hopefully import errors
    deter people away from this.

    ex:
    file_name = 'flask.router.Router'
    active_library_names = ['flask', 'wsgi', 'sqlalchemy', 'git']
    returns False

    file_name = 'core.utils.docker_constants'
    active_library_names = ['flask', 'wsgi', 'sqlalchemy', 'git']
    return True

    :param file_name: string of file_name without file type (contrast.utils.stack_trace_utils)
    :return: True if no active libraries are in the file_name else False
    """
    active_libraries_names = [
        # exclude vulnpy - a purposely-vulnerable library
        lib
        for lib in get_active_library_names_from_pkg()
        if lib != "vulnpy"
    ]
    for library_name in active_libraries_names:
        if file_name.startswith(library_name):
            return False
    return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.agent import agent_state
from contrast.agent.middlewares.route_coverage.common import build_key

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class RoutesMixin(object):
    """
    Class for route coverage work.
    Route coverage is assess only.
    """

    def handle_routes(self, context, request):
        """
        Method that should run for all middlewares immediately after the application
        executes.
        """
        if context and request:
            self.get_routes_on_first_request()
            self.check_for_new_routes(context, request)
            self.append_route_to_findings(context)

    def get_routes_on_first_request(self):
        """
        Attempt to do route discovery on the first request
        """

        # if we already have routes here, we must have populated them from a framework-
        # specific patch. Eventually, `get_routes_on_first_request` will be completely
        # removed, once all frameworks perform route discovery using patches.
        if not agent_state.is_first_request():
            logger.debug("skipping route discovery - not the first request")
            return
        if agent_state.get_routes():
            logger.debug("skipping route discovery - already populated")
            return

        logger.debug("First request so looking for app routes.")

        discovered_routes = self.get_route_coverage()

        if not discovered_routes or not isinstance(discovered_routes, dict):
            logger.debug("Could not find any routes")
            return

        routes = agent_state.get_routes()
        routes.update(discovered_routes)

        routes_to_log = [f"{route.verb}: {route.url}" for route in routes.values()]
        logger.debug("Found the following routes: %s", routes_to_log)

    def check_for_new_routes(self, context, request):
        """
        Check to see if the current request is a new route. If so, add this to the list
        of all discovered routes.
        """
        logger.debug("Checking for new route.")

        if context.view_func is None:
            context.view_func = self.get_view_func(request)
            if context.view_func is None:
                logger.debug(
                    "unable to determine view function for the current request"
                )
                return
        else:
            logger.debug("already have the view function for this request")

        request_method = request.method
        route_id = build_key(str(id(context.view_func)), request_method)

        self.update_route_information(context, route_id, request_method)

    def update_route_information(self, context, route_id, request_method):
        """
        Given a context and route_id, check if the route_id is in middleware (self)
        routes. Store the route as current and observed route for later use.

        :param context: RequestContext instance
        :param route_id: string id for a route
        :param request_method: string such as 'GET'
        :return: no return, side effects only
        """
        routes = agent_state.get_routes()
        if route_id not in routes:
            url = context.request.get_normalized_uri()

            if context.view_func_str is None:
                context.view_func_str = self.build_route(context.view_func, url)

            routes[route_id] = Route(
                verb=request_method, url=url, route=context.view_func_str
            )

        route = routes[route_id]
        logger.debug("Route visited: %s : %s", route.verb, route.url)

        context.current_route = route

        # Currently we do not report an observed route if the route signature is empty.
        # As a team we've decided there isn't a meaningful default signature value
        # we can provide to customers. If a route doesn't show up in Contrast UI,
        # it may be due to its missing signature. In this scenario, we will have to work
        # with the customer directly to understand why the signature was not created.
        if not route.signature:
            logger.debug(
                "No route signature found for %s : %s (id=%s). Not updating observed"
                " route",
                route.verb,
                route.url,
                route_id,
            )
            return

        context.observed_route.signature = route.signature

        context.observed_route.url = route.url
        context.observed_route.verb = route.verb

    def append_route_to_findings(self, context):
        """
        Route discovery and current route is not identified until the handle ensure
        part of the request lifecycle, after assess has analyzed and potentially created
        a finding, so that is why we have to append the now-available current route to
        the existing finding.
        """
        if not context.current_route:
            logger.debug("No current route to append to findings")
            return

        if not context.findings:
            logger.debug("No findings to append route to")
            return

        for finding in context.findings:
            if not finding.routes:
                logger.debug(
                    "Appending route %s:%s to %s",
                    context.current_route.verb,
                    context.current_route.url,
                    finding.rule_id,
                )
                finding.routes.append(context.current_route)

            finding.set_version()

    def get_view_func(self, request):
        return

    def build_route(self, view_func, url):
        return ""



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from copy import copy
from contrast.agent.assess.tag import (
    ABOVE,
    BELOW,
    HIGH_SPAN,
    LOW_SPAN,
    WITHIN,
    WITHOUT,
    Tag,
)


def intersection(tags, tag_ranges):
    """
    Return a list of tags trimmed down to the specified tag ranges.
    If a single tag has intersections with several ranges, the tag is split
    into the appropriate number of new tags.

    To clarify, for each tag in tags, we find its intersection with each
    tag in tag_ranges and combine those results.
    """
    if not tags:
        return []

    result = []
    for tag in tags:
        parts = []
        if tag.length <= 0:
            continue
        for tag_range in tag_ranges:
            new_tag = copy(tag)
            if new_tag.intersect(tag_range):
                parts.append(new_tag)
        if len(parts) > 0:
            result.extend(combine(parts))

    return result


def is_covered(tag_ranges, covering_ranges):
    """
    Computes whether `tag_ranges` are completely covered by `covering_ranges`
    """
    if not tag_ranges:
        return True

    tag = tag_ranges[0]

    for item in covering_ranges:
        if tag.length <= 0:
            break

        comparison = tag.compare_range(item.start_index, item.end_index)

        if comparison in [BELOW, LOW_SPAN, WITHOUT]:
            return False

        if comparison == WITHIN:
            tag = Tag(0, 0)
        elif comparison == HIGH_SPAN:
            tag = Tag(tag.end_index - item.end_index, item.end_index)
        elif comparison == ABOVE:
            continue

    if tag.length > 0:
        return False

    return is_covered(tag_ranges[1:], covering_ranges)


def ordered_merge(tag_array, new_tag_array):
    if isinstance(new_tag_array, list):
        if not new_tag_array:
            return tag_array

        if not tag_array:
            return new_tag_array

        for new_item in new_tag_array:
            _ordered_merge(tag_array, new_item)

    else:
        if not new_tag_array:
            return tag_array

        if not tag_array:
            return [new_tag_array]

        _ordered_merge(tag_array, new_tag_array)

    return combine(tag_array)


def merge_tags(tags):
    if isinstance(tags, dict):
        for key, value in tags.items():
            tags[key] = combine(value)


def _ordered_merge(tag_array, new_item):
    index = 0

    for existing in tag_array:
        if existing.start_index >= new_item.start_index:
            break

        if existing.overlaps(new_item):
            existing.merge(new_item)
            return

        index = index + 1

    tag_array.insert(index, new_item)


def combine(tags):
    if not tags:
        return tags

    current = tags[0]
    combined = [current]

    for tag in tags[1:]:
        if tag.start_index <= current.end_index:
            if tag.end_index > current.end_index:
                current.set_end(tag.end_index)
        else:
            current = tag
            combined.append(current)

    tags = combined

    return tags



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import Enum, auto

import contrast
from contrast.agent.request import Request
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ProtectResponse(Enum):
    NO_ACTION = auto()
    BLOCKED = auto()
    MONITORED = auto()
    PROBED = auto()
    BLOCKED_AT_PERIMETER = auto()
    SUSPICIOUS = auto()


_RESPONSE_TO_TS_NAME = {
    ProtectResponse.MONITORED: "exploited",
    ProtectResponse.BLOCKED: "blocked",
    ProtectResponse.PROBED: "ineffective",
    ProtectResponse.SUSPICIOUS: "suspicious",
}


class Attack:
    """
    Class storing all data necessary to report a protect attack.
    """

    def __init__(self, rule_id):
        self.rule_id = rule_id
        self.samples = []
        self.response = None
        self.start_time_ms = contrast.CS__CONTEXT_TRACKER.current().request.timestamp_ms
        self.response = None

    @property
    def blocked(self):
        return self.response == ProtectResponse.BLOCKED

    def add_sample(self, sample):
        self.samples.append(sample)

    def set_response(self, response):
        self.response = response

    def _convert_samples(self, request: Request):
        if request is not None:
            reportable_request = request.reportable_format
        else:
            reportable_request = {}

        return [
            {
                "blocked": self.blocked,
                "input": {
                    "documentPath": sample.user_input.path,
                    "documentType": sample.user_input.document_type.name,
                    "filters": sample.user_input.matcher_ids,
                    "name": sample.user_input.key,
                    "time": sample.timestamp_ms,
                    "type": sample.user_input.input_type.name,
                    "value": sample.user_input.value,
                },
                "details": sample.details,
                "request": reportable_request,
                "stack": [
                    {
                        "declaringClass": stack.declaring_class,
                        "methodName": stack.method_name,
                        "fileName": stack.file_name,
                        "lineNumber": stack.line_number,
                    }
                    for stack in sample.stack_trace_elements
                ],
                "timestamp": {
                    "start": sample.timestamp_ms,  # in ms
                    "elapsed": (
                        now_ms() - sample.timestamp_ms
                    ),  # in ms which is the format TS accepts
                },
            }
            for sample in self.samples
        ]

    def to_json(self, request: Request):
        common_fields = {
            "startTime": 0,
            "total": 0,
        }
        json = {
            "startTime": self.start_time_ms,
            "blocked": common_fields,
            "exploited": common_fields,
            "ineffective": common_fields,
            "suspicious": common_fields,
        }

        relevant_mode = _RESPONSE_TO_TS_NAME.get(self.response)
        if relevant_mode is None:
            # Don't know what response is so just report default info so we can debug.
            return json

        samples = self._convert_samples(request)

        json[relevant_mode] = {
            "total": 1,  # always 1 until batching happens
            "startTime": self.start_time_ms,
            "samples": samples,
        }

        return json



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast_vendor import wrapt

import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy import registry
from contrast.agent.middlewares.route_coverage.starlette_routes import (
    create_starlette_routes,
)
from contrast.agent.policy import patch_manager
from contrast.agent.middlewares.route_coverage.common import build_route
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


STARLETTE_ROUTING = "starlette.routing"
STARLETTE_REQUESTS = "starlette.requests"
SESSION = "session"


def build___call___patch(orig_func, patch_policy):
    """
    Patch for starlette.routing.Router.__call__

    This currently gives us route discovery and route observation. If in the future we
    need a reference to the actual FastAPI / Starlette application object, we'll need
    another patch (probably for one of those class's __call__ methods).

    The innermost application object in any Starlette-based application is a Router.
    The router is a valid ASGI application object. We cannot patch
    starlette.applications.Starlette.__call__ (or fastapi.FastAPI.__call__) because of
    the order in which these methods are invoked during request processing. Starlette's
    unique middleware installation procedure leads to the following call order:

    fastapi.FastAPI.__call__
      starlette.applications.Starlette.__call__
        each middleware's __call__
          starlette.routing.Router.__call__

    This differs from typical middleware installation, where the middleware __call__
    methods come first.

    We need to perform route discovery / observation while we're still somewhere inside
    of the agent middleware's __call__ method; otherwise we won't have this information
    in time for handle_ensure() at the end of agent request processing. The best option
    seems to be Router.__call__.
    """
    del patch_policy

    async def __call___patch(wrapped, instance, args, kwargs):
        try:
            result = await wrapped(*args, **kwargs)
        finally:
            do_starlette_route_discovery(instance)
            do_starlette_route_observation(instance, *args, **kwargs)
        return result

    return wrap_and_watermark(orig_func, __call___patch)


@fail_quietly("Failed to run starlette first-request analysis")
@scope.with_contrast_scope
def do_starlette_route_discovery(starlette_router_instance):
    from contrast.agent import agent_state

    if not agent_state.is_first_request():
        return

    discovered_routes = create_starlette_routes(starlette_router_instance)

    agent_state.update_routes(discovered_routes)
    logger.debug(
        "Discovered the following starlette routes: %s",
        [f"{route.verb} {route.url}" for route in discovered_routes.values()],
    )


@fail_quietly("unable to perform starlette route observation")
@scope.with_contrast_scope
def do_starlette_route_observation(starlette_router_instance, *args, **kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    logger.debug("Performing starlette route observation")

    asgi_scope = args[0] if len(args) > 0 else kwargs.get("scope")
    if not asgi_scope or not isinstance(asgi_scope, dict):
        logger.debug(
            "unable to get ASGI scope for route observation. args: %s, kwargs: %s",
            args,
            kwargs,
        )
        return

    view_func = asgi_scope.get("endpoint")
    if view_func is None:
        logger.debug("did not find endpoint for starlette route observation")
        return

    context.view_func = view_func
    context.view_func_str = build_route(view_func.__name__, view_func)
    logger.debug("Observed starlette route: %s", context.view_func_str)


class ContrastSessionDictProxy(wrapt.ObjectProxy):
    """
    Custom ObjectProxy we use to wrap dicts returned by starlette's request.session
    property. These proxied dicts have a trigger for trust-boundary-violation on
    __setitem__.
    """

    def __setitem__(self, key, value):
        result = None
        try:
            result = self.__wrapped__.__setitem__(key, value)
        finally:
            analyze_setitem(result, (self, key, value))

        return result


@fail_quietly("Failed to analyze session dict __setitem__")
def analyze_setitem(result, args):
    context = contrast.CS__CONTEXT_TRACKER.current()
    policy = registry.get_policy_by_name("starlette.sessions.dict.__setitem__")
    analyze(context, policy, result, args, {})


def build_session_patch(orig_prop, patch_policy):
    def session_fget(*args, **kwargs):
        """
        Function used to replace fget for starlette's request.session property.
        This function returns proxied dictionaries - see ContrastSessionDictProxy.
        """
        session_dict = orig_prop.fget(*args, **kwargs)

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            return session_dict

        return ContrastSessionDictProxy(session_dict)

    return property(session_fget, orig_prop.fset, orig_prop.fdel)


def patch_starlette_requests(starlette_requests_module):
    build_and_apply_patch(
        starlette_requests_module.Request, SESSION, build_session_patch
    )


def patch_starlette_routing(starlette_routing_module):
    build_and_apply_patch(
        starlette_routing_module.Router, "__call__", build___call___patch
    )


def reverse_patches():
    starlette_routing_module = sys.modules.get(STARLETTE_ROUTING)
    if starlette_routing_module:
        patch_manager.reverse_patches_by_owner(starlette_routing_module.Router)

    starlette_requests_module = sys.modules.get(STARLETTE_REQUESTS)
    if starlette_requests_module:
        patch_manager.reverse_patches_by_owner(starlette_requests_module.Request)


def register_patches():
    wrapt.register_post_import_hook(patch_starlette_requests, STARLETTE_REQUESTS)
    wrapt.register_post_import_hook(patch_starlette_routing, STARLETTE_ROUTING)
